#include "include.h"
#include "idrp.h"

/* idrp_rt_peer - 
 * routing routines to handle peers up/down
 * and refresh request cycles.
 *
 * Also included in this is the Adjacency Rib
 * check routines
 * 
 *  Rib handling routine
 *
 * void consistency_check (rib_no) (John)
 *      int rib_no - Rib ID of attributes
 *      -1 - all ribs
 *      now only 0 - default
 *
 * void validate_AdjRib(rib_id,peer) (John)
 *	rib_id - validate for rib_id n 
 *      idrpPeer *peer  - peer structure to check
 *                        adjacency Rib in
 *      run checksum on data - to match specification
 * 	minimum 
 *  			
 * void validate_allRibs()
 *     validate all Ribstructures with a 
 *     idrp appropriate checksum
 * 
 *  %% Full routing blast routines  for IDRP
 *  
 * idrp_rt_send_init_routes(peer) -
 *     send initial blast of routes to neighbor
 *       creates a send list for this peer by:
 *	 - get all active routes from gated
 *       - create send list
 *  	 hands send list to dist_policy
 *	 hands modified send list to send_updates
 *		which then sends pdus out one by one 
 *              via send_update_pdu in idrp.c
 * 
 * idrp_peer_down(peer) - delete all routes for Peer
 * 			 from gated, phase 1 processing  
 * 			 to replicate routes out
 *
 * idrp_rt_re-init -
 * 		re-init after policy change to
 *		drop and re-establish peers
 *  
 *  idrp_send_refresh_all_routes(peer) 
 *	simply calls the above routine  
 *
 * path_up - sets path up, starts timers,
 *	     calls idrp_rt_init
 * 
 * 
 * idrp_refresh_all_peers -
 *  - gets announce list and 
 *    sends peer refresh sequence
 * 
 *
*/ 

/* %% AdjRib checks */

/*
 * Check the route table for consistency
 */
static void consistency_check(rib_no)
int rib_no;
{
	/* idea behind the routine */
	/* if rib_no = -1 - check all IDRP routes */
	/* if rib_no = 0 - check active default   */

	/* do do consistency check */
 
	/* 1.) walk the gate table and check the links between */
	/*      nlri for one route_id */
	/* 2.) walk the outbound route_out id arrays at same time */

	/* define the consistency error to be one so that we */
	/* can see that it works */
 
	consistency_errors = 0;

	if (consistency_errors) {
		trace(TR_ALL, LOG_ERR, "errors found in rib att %s",rib_no);
		kill(task_pid, SIGINT);	/* Take a dump */
	}
}

 /* 			
 * 2.) void validate_AdjRib(peer) (John)
 *      idrpPeer *peer  - peer structure to check
 *                        adjacency Rib in
 */
void 
validate_AdjRib(rib_id,peer)
int rib_id;
idrpPeer *peer;
{
	/* call the rt_locate with gateway to make sure */
	/* Adj_Rib in is check */
	/* save a checksum for this gateway in peer structure */
	/* in checksum in Peer structure */
	/* at first do a TCP checksum on all NLRI for now */
	/* other at John's decision */

	return;
}


/* 3.) void validate_allRibs()
 *     validate all Ribstructures with a 
 *     idrp appropriate checksum
 */
void
validate_allRibs(rib_no)
int	rib_no;
{
idrpPeer *peer;

	IDRP_LIST(peer)
		{
		validate_AdjRib(rib_no, peer);
		}
	IDRP_LIST_END
	return;
}


/* %% Full routing dump routines
 * 
 *  idrp_rt_send_init -
 * 	(called by path_up)
 * 
 *  
 * idrp_send_refresh_all_routes -
 *    send unsolicity refresh request 
 *    this must be called by a kill -USR1 
 * 
 *  peer_down - pull all routes for peer 
 * idrp_rt_re_init - re-init after policy
 *		     change 
 *
 */ 

/* 
 *idrp_rt_send_init - 
 * Send the entire LocRib to a peer.
 * upon initial bring up of routes
 */

void idrp_rt_send_init(peer)
idrpPeer *peer;
{
rt_list *rtl[IDRP_NLRI_FAMILIES_SUPPORTED];
rt_entry *route;
u_int proto;
int	i;
int	error;

	NLRI_FAMILY_LOOP(i)
		{
		if (!peer->nlri_supp[i])
			{
			rtl[i] = (rt_list *) 0;
			continue;
			}
	
		/* protocol supported - so try to get active list */
	
		rtl[i] = rthlist_active(peer->nlri_supp[i]);
		} NLRI_FAMILY_LOOP_END 
		  
	idrp_phase3_dump(peer,rtl);
}

void
idrp_send_refresh_all_routes(peer)
idrpPeer *peer;
{
	/* send all Loc_Rib to this peer null for now */ 
	/*  - send rib refresh start at beginning
	 * -  call idrp_rt_send_init
	 * - send rib refresh end 
	 */

	trace (TR_IDRP,0,"send full refresh of active routes to peer ");
}


/*
 * A peer has gone down.  Delete its routes.
 */

void idrp_peer_down(peer)
idrpPeer *peer;
{
/* gated list pointers for phase 1 handling */

int 	changes = 0;
idrpRoute	*p_idrp_rt;
idrp_ann_list	ann_list;
rt_list	*rtl;


	bzero(&ann_list,sizeof(idrp_ann_list));
	ann_list.rib_id = RIB_ID_DEFAULT;
	ann_list.peer = peer;

	/* note the IP routes and the ISO routes
	 * are called explicitly here
	 * - if we go to other protocols,
	 *   we will need to change this
	 *  to an NLRI_FAMILY_LOOP
	 */


	rt_open(peer->task);

	/* get the IP routes for this peer */

	rtl = rtlist_gw(&peer->gw);

	/* gather them up */
 
	changes = idrp_peer_route_pull(rtl,&ann_list);

	/* get the ISO routes for this peer */

	rtl = rtlist_gw(&peer->iso_gw);

	/* gather them up */

	changes += idrp_peer_route_pull(rtl,&ann_list);

	/* now process the announce list in one fell swoop
	 * - if it is internal peer, just delete
	 * - if it is external peer, you may need
	 *   to send out best_external routes and
	 *   then complete delete of routes from peer 
	 */
 
	 
	
	if (peer->type == IDRP_PEER_INTERNAL)
		{
		phase1_internal(&ann_list);
		}
	else
		{
		phase1_external(&ann_list);
		}

	/* Now clean out all of the route ID hash table
	 * entries for this peer.
	 */
	idrp_reinit_hash_tbl(peer);
	
	rt_close(peer->task,&peer->gw,changes, NULL);
}


idrp_peer_route_pull(rtl,p_ann_list)
rt_list 	*rtl;
idrp_ann_list	*p_ann_list;
{
register rt_entry	*rt = (rt_entry *) 0;	 
idrpRoute		*p_idrp_rt;
int			nlri_id;
int			changes = 0;

	/* do my own delete so I can send phase1
	 * - withdrawls to internal neighbors
	 * - normally would use rt_gwunreach call
	 * - in gated
	 */

	RT_LIST(rt,rtl,rt_entry)   {
		/* walk chain of this destination
	 	 * all these routes are idrp by definition
		 *  1.) link the withdraw by  
		 *   
		 */ 

		p_idrp_rt = (idrpRoute *) rt->rt_data;
		link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_WITH);
		changes++;
		} RT_LIST_END(rt,rtl,rt_entry);
	return(changes);
}

/* idrp_rt_reinit - redo all policy after a reinit.  The things which can
 *     change which effect us are the ?? - for now just return
 */

void
idrp_rt_reinit(peer)
idrpPeer *peer;
{
}
/*
 * Bring up the path to our neighbor
 */
void path_up(peer)
idrpPeer *peer;
{
	/* increment the number of idrp peers */

	idrp_n_established++;
	peer->idrp_flags = IDRPF_CONNECT;	
	trace(TR_ALL, LOG_ERR, "IDRP path to %s is up", peer->name);
	
	send_keepalive(peer);
	start_keepalive_timer(peer);
#ifdef ECHOKLUDGE
	start_echo_timer(peer);
#endif
	idrp_rt_send_init(peer);
}


void
idrp_refresh_all_peers(p_ann_list)
idrp_ann_list	*p_ann_list;
{
idrpPeer	*peer;
idrp_ann_list	*p_atl;
int		i;
idrpRoute	*p_irt;
idrp_send_list	*p_send_list = 0;

	IDRP_LIST(peer)
		{
		if (peer->idrp_flags != IDRPF_CONNECT)
			{
			continue;
			}

		send_Refresh_pdu(peer,IDRP_RIB_REFRESH_START);

		/* build and send the updates for the Rib Refresh peer
		 * one thing that can be done in the future is
		 * to hand the update code a linked send list
		 * - right now we just do one at a time
		 * and loop here
		 */ 


		ANN_LIST(p_atl,p_ann_list)
			{
			/* walk NLRI list for this attribute */	
			NLRI_FAMILY_LOOP(i)
				{
				/* walk the announce tree for nlri this attribute */
				ANN_ANN_LIST_WALK(p_irt,i,p_atl)
					{	
					/* link attribute to send list if OK for peer */
				
					if (DIST(p_irt,peer,IDRP_DIST_REFRESH))
						{	
						p_send_list = send_nlri_attr(p_irt,p_atl->p_attr,p_send_list,peer);
						}
					} ANN_ANN_LIST_WALK_END;
				} NLRI_FAMILY_LOOP_END;
			flush_att_send_list(peer,p_atl->p_attr,p_send_list);
			} ANN_LIST_END;		
		
		free_send_list(p_send_list,peer);
		send_Refresh_pdu(peer,IDRP_RIB_REFRESH_END);

		/* queue a keepalive */
		start_keepalive_timer(peer);
	} IDRP_LIST_END;
			
}		

void
idrp_req_refresh(peer)
idrpPeer	*peer;
{

	/* requesting Rib Refresh from peer from gated NM action
	 * or from internal determination that Rib is not good
	 */

	trace (TR_IDRP,0,"request rib refresh from peer %s",peer->name);
	send_Refresh_pdu(peer,IDRP_RIB_REFRESH_REQ);
}
