#include "include.h"
#include "idrp.h"

/* idrp_rt_phase1. c 
 *  Routing table code for IDRP - phase 1 routing routines 
 * 
 * 
 * IDRP decision routines  
 *
 * Phase 1 decision routes -
 * 
 *  idrp_pref(p_idrp_rt) - calculate preference
 *     for route based on the peer
 * 
 *  idrp_ext_info_pref(p_rt) - calculate idrp preference
 *				on ext_info 
 *  
 *  
 * %% Phase 1 processing
 * %% processing of update BISPDU routines
 * 
 * idrp_route_mod - add route to update parse change list
 * idrp_process_pdu_routes - process routes after parse
 * 
 *  
 * %% Phase one processing - per idrp spec
 * 
 * idrp_phase1 _process - process phase1 
 * phase1_internal 	- handling of internal routes in phase 1
 * phase1_external 	- handling of external routes in phase 1
 * ph1_with_int_route 	- phase1 withdraw internal route
 * ph1_with_rep_int_route - phase1 withdraw/replace internal route
 * ph1_add_int_route 	- phase 1 add internal route
 * ph1_with_ext_route 	- phase 1 withdrawl of external route
 * ph1_with_rep_ext_route - phase1 - withdraw with replace of external route
 * ph1_add_ext_route 	- phase1 add external route
 * send_best_ext 	- send best external routes
 * 
 * del_route_best_ext	- delete best external routes withdrawn
 * remove_ann_dup	- remove announce duplicates from announce list
 *  
 *    
 *
 * We have the following  entry points to the IDRP code:
 * idrp_process_update -> parse_update -> idrp_route_mod
 *			 and when done -> idrp_process_pdu_routes
 * 
 * idrp_process_pdu_routes -> calls phase 1 processing routines:
 * -> phase1_internal->ph1_int_add_route 
 *		     ->ph1_with_route
 *		     ->ph1_with_rep_route
 * -> phase1_external->ph1_add_ext_route
 *		     ->ph1_with_route
 *		     ->ph1_with_rep_route
 *	-> send_best_ext -> del_routes_best_ext
 *   
 * Before the routes are handed to gated, preferences for
 *  gated are set so that it can proform phase 2 processing:
 *    check for other route
 *    tie_break routine
 *    gated_pref routine 
 *   
 *  
 * idrp_flash - called to handle the list of new active routes
 * 		from the gated routing table.
 *              The phase 3 routes are only sent to external peers.
 * 
 * idrp_flash -> phase3_processing -> phase3_status_change
 * ->idrp_send_phase3_route -> idrp_del_ph3_routes
 * 
 * 
 *
 * Yet to work on for idrp - new policy and re-init routines
 * auxiliary policy and and terminate gateway routines :
 *
 * names will be something like:
 * idrp_rt_reinit
 * idrp_rt_terminate
 * idrp_newpolicy  
 * idrp_aux_policy(). 
 *
 */ 


 /*  %%  IDRP decision routines   */

 /*
 * Phase 1 decision routes -
 * 
 *  idrp_pref(p_idrp_rt,peer) - calculate preference
 *     for route based on the peer
 */

void idrp_pref(p_idrp_rt)
idrpRoute *p_idrp_rt;
{
	/* right now preference is easy - just give hop count
	 * in the future call PREF - which will do import policy on
	 * idrp route, given idrp policy structure
	 */

	p_idrp_rt->pref_cal = p_idrp_rt->p_attr->hopcount;
		
	
return;
} 

/* idrp_ext_info_pref (rt_entry *p_rt)
 * 		ext_info preference for idrp
 *
 */
 
int
idrp_ext_info_pref(p_rt)
rt_entry	*p_rt;
{
proto_t	proto;
int	family;
int	pref;

	/*  Need to check whether route is valid to
	 *  import into IDRP, and if this is an
	 *  EXT_INFO route or not 
	 * 
	 *  Preference will be zero if this
	 *  is not be imported to IDRP
	 *
	 *  Preference setting of IDRP_EXT_INFO_IGP  if
	 *  this route is internal IGP route
	 * 	(for now IS-IS needs to be
	 *	 having no RTS_EXTERNAL
	 *	 flag set for gated.  This
	 *	 should translate into L2 internal
	 *	 routes and no Code 5 routes (page 60
	 *	 IS-IS specification) 
	 *	 
	 * Preference of IDRP_EXT_INFO_EXT_INFO this value
	 * will indicate an external info needing
	 * the idrp protocol EXT_INFO
	 */ 
		

	proto = p_rt->rt_gwp->gw_proto;
	switch (proto)
		{
		case RTPROTO_DIRECT:
			return(0);
			break;

		case RTPROTO_KERNEL:
			return(0);
			break;
		case RTPROTO_REDIRECT:
			return(0);
			break;
		case RTPROTO_DEFAULT:
			return(IDRP_EXT_INFO_IGP);
			break;

		case RTPROTO_OSPF:
			/* no translation for OSPF
			 * until routines are written
		 	 * to implement the BGP4/IDRP to OSPF
			 * specification - No half baked solution
			 * here
			 */
 		
			return(0);
			break;

		case RTPROTO_OSPF_ASE:
			/* no translation for OSPF
			 * until routines are written
		 	 * to implement the BGP4/IDRP to OSPF
			 * specification - No half baked solution
			 * here
			 */
			return(0);
			break;

		case RTPROTO_IGRP:
			/* IGRP - no transtlation as of yet */
 
			return(0);
			break;

		case RTPROTO_RIP:
			/* RIP - no translation as of yet */
			return(0);
			break;
			
		case RTPROTO_BGP:
			/* BGP - no translation as of yet */
			return(0);
			break;

		case RTPROTO_EGP:
			/* EGP- no translation as of yet */
			return(0);
			break;
			
		case RTPROTO_STATIC:
			if (p_rt->rt_state & RTS_EXTERIOR)
				{
				return (IDRP_EXT_INFO_EXT_INFO);
				}
			else
				{
				return(IDRP_EXT_INFO_IGP);
				}
			break;
			
		case RTPROTO_SNMP:
			/* SNMP route - not now
			 * - but later we try static style
			 * routes 
			 */
	 
			return(0);
			break;

		case RTPROTO_IDPR:
			/* IDPR - needs specification of
			 * what is normal here
			 */
 
			return(0);
			break;

		case RTPROTO_ISIS:
			/* IS-IS routes
			 * - need to test RTS_EXTERIOR
		 	 *   if set EXT_INFO route
			 *   if not set, IGP route
			 */
 
			if (p_rt->rt_state & RTS_EXTERIOR)
				{
				/* check to see that it is not
				 * my bis 
				 */
				pref = 0;
				if (!my_bis_peer(p_rt->rt_gwp))
					{
					/* not my peer - so announce the route */
					pref = IDRP_EXT_INFO_EXT_INFO;
					}
				}
			else
				{
				pref = IDRP_EXT_INFO_IGP;
				}				
			
			return(pref);
			break;

		case RTPROTO_SLSP:
			return(0);
			break;

		case RTPROTO_IDRP:
			trace (TR_IDRP,0,"Illegal protocol interactions - IDRP should not call ext_info pref");
			return (0);
			break;
 
		case RTPROTO_INET:
			return(0);
			break;
		
		case RTPROTO_IGMP:
			return(0);
			break;

		}
	return (0);

}


/*  %% phase 1 processing */
 

/*
 * idrp_process_pdu_routes(peer,res)
 * Process the routes received in a pdu.  
 */

idrp_ann_list *
idrp_process_pdu_routes(peer,res,type)
idrpPeer	*peer;
parse_results	*res;
int		type;
{
/* general pointers for processing */

idrp_rid_hash	*p_rid;			/* hash table entry */
#ifdef ADN_PATCH
idrpRoute_link	with_list[IDRP_NLRI_FAMILIES_SUPPORTED];
#else
idrpRoute_link	with_list[IDRP_NLRI_FAMILIES_SUPPORTED] = {0};
					/* pointer to list of nlri's to withdraw */
#endif /* ADN_PATCH */
					/* pointer to list of nlri's to withdraw */
idrpRoute	*p_rt;			/* pointer to idrp route structure */
idrp_ann_list	*p_ann_list;		/* announce list structure */
idrpRoute	*p_last_irt = 0;	/* loop holder for tail of ann_nlri list */ 

/* loop counters */

int		i;
int		j;
int		k;

#ifdef ADN_PATCH
	{
		register int cnt;
		for (cnt=0; cnt < IDRP_NLRI_FAMILIES_SUPPORTED; cnt++) {
			with_list[cnt].head = (struct  _idrpRoute *) 0;
			with_list[cnt].tail = (struct  _idrpRoute *) 0;
		}
	}
#endif

 
    /* process the routes from one pdu
     * withdrawal routes 1st
     * allocate and initialize announce lsit
     * with array initialized above
     */ 

    p_ann_list = (idrp_ann_list *) task_mem_malloc(peer->task,sizeof(idrp_ann_list));
    bzero(p_ann_list,sizeof(idrp_ann_list)); 	

    if (res->withdraw->count)
	{
	for (i = 0; i < res->withdraw->count ;i++) 
		{
		/* hash the route id to get first route pointer 
		 * route - rt_entry pointer -> rt_data and idrp_Route info 
		 * idrp_Route_info -> other nlri associated with route_id 
		 */

		p_rid = idrp_rid_hash_get(peer,res->withdraw->withdraw[i]);
		if (p_rid)
			{
			/* found hash - get rid of any nlris in list that 
			 *  are being announced by pdu
			 */ 

			NLRI_FAMILY_LOOP(j)
				{
				/* found the route - now go about creating withdraw nlri 
		 		 * list from this withdraw id plus the announce nlri. 
				 * This takes care of withdrawal id (10 nlri) - reannounce 
				 * 8 of the nlri so we can replace instead of w/drawing 
				 * and then immediately re-announcing.
				 */

				/* do we have a withdraw list started for this NLRI  */
				if (with_list[j].tail)
					{
					/* withdraw list exists - so add to it */
					p_rt =  remove_ann_dup(p_rid->p_idrp_rt[j],res->rt_chg[j]);
					with_list[j].tail->p_with = p_rt;
					}
				else
					{
					with_list[j].head = remove_ann_dup(p_rid->p_idrp_rt[j],res->rt_chg[j]); 
					with_list[j].tail = with_list[j].head; 
					}
				} NLRI_FAMILY_LOOP_END
			} /* end of if (p_rid) */

		/* release the hash table entry now that the entry is done */

		idrp_free_hash_entry(peer,res->withdraw->withdraw[i]);	

		} /* end of for loop for each withdraw route_id */

	    /* save withdraw nlri list in announce list */
		NLRI_FAMILY_LOOP(i)
			{ 
			p_ann_list->with_nlri[i].head = with_list[i].head;
			p_ann_list->with_nlri[i].tail = with_list[i].tail;
			} NLRI_FAMILY_LOOP_END

	    } /* end of withdraw nlri processing */
	else
            {
	    /* no withdraw nlri list */

	     NLRI_FAMILY_LOOP(i)
		{ 
		p_ann_list->with_nlri[i].head = (idrpRoute *) 0; 
		p_ann_list->with_nlri[i].tail = (idrpRoute *) 0; 
		} NLRI_FAMILY_LOOP_END
	    } 

	/* set-up announce list to pass to phase 1 processing */

	p_ann_list->peer = peer;
	p_ann_list->p_next = 0;
	p_ann_list->p_attr = res->p_att;	

	/* for now set the rib_id to default */
	p_ann_list->rib_id = res->p_att->rib_id; 

	NLRI_FAMILY_LOOP(i)	
		{
		/* set up head and tail of ann_nlri list
		 * linked by p_next_nlri
		 */

		p_ann_list->ann_nlri[i].head = res->rt_chg[i];
		p_last_irt = 0;
		for (p_rt = res->rt_chg[i]; p_rt; p_rt = p_rt->p_next_nlri)
			{
			p_last_irt = p_rt;
			}	
		p_ann_list->ann_nlri[i].tail = p_last_irt;
		} NLRI_FAMILY_LOOP_END

	/* set-up the hash table entry for the route_id and 
	 * route_id list 
	 */

	idrp_add_hash_entry(peer,res->route_id_in,res->rt_chg);

	if (type == IDRP_RIB_REFRESH_TYPE)
		{
		return (p_ann_list);
		}
	else
		{
		idrp_phase1_process(p_ann_list);
		task_mem_free(peer->task,p_ann_list);
		return (p_ann_list);
		}

}	

void idrp_phase1_process(ann_list)
idrp_ann_list 	*ann_list;
{
int	changes = 0;
/* Phase 1 logic - 
 * splits on internal or external
 */
	rt_open(ann_list->peer->task);
	switch (ann_list->peer->type)
		{	
		case IDRP_PEER_INTERNAL:
			trace (TR_IDRP | TR_UPDATE,0, "peer %s phase1 internal pdu processing ",ann_list->peer->name);

			changes = phase1_internal(ann_list);
			break;
	
		case IDRP_PEER_EXTERNAL:
			trace (TR_IDRP | TR_UPDATE,0,"peer %s phase1 external pdu processing",ann_list->peer->name);
			changes = phase1_external(ann_list);
			break;
		}
	/* close gated route table */

	rt_close(ann_list->peer->task,&ann_list->peer->gw,changes,NULL);
}

int
phase1_internal(ann_list)
idrp_ann_list	*ann_list;
{
idrp_ann_list	*p_atl;
idrpPeer 	*peer;
idrpRoute	*p_rt,*p_rt_next;
int		i;
int		changes = 0;
	
	/* process nlri families separately */

	ANN_LIST(p_atl,ann_list)
		{
		NLRI_FAMILY_LOOP(i) 
			{
			/* process withdrawals */

			ANN_WITH_LIST_DEL_WALK(p_rt,i,p_atl,p_rt_next)
				{
				/* walk through withdrawal list */
				p_rt_next = p_rt->p_with;
				ph1_with_int_route(p_rt);
				changes++;
				}
		
			/* process announces */
			ANN_ANN_NEXT_NLRI(p_rt,i,p_atl)
				{
				if (p_rt->status & IDRP_STATUS_REPLACE)
					{
					/* handle replaces */
					ph1_with_rep_int_route(p_rt);
					}
				else
					{
					/* handle additions */
					ph1_add_int_route(p_rt);
					}
				changes++;
				} ANN_ANN_NEXT_NLRI_END;
			}  NLRI_FAMILY_LOOP_END
		} ANN_LIST_END;
	
	return(changes);
}


int
phase1_external(ann_list)
idrp_ann_list	*ann_list;
{
idrp_ann_list	*p_atl;
idrpRoute	*p_rt,*p_rt_next;
idrp_ann_list	out_ann_list;
int		i;
int		changes = 0;	
	/* process nlri families separately */

	bzero(&out_ann_list,sizeof(idrp_ann_list));


	ANN_LIST(p_atl,ann_list)
		{	
		NLRI_FAMILY_LOOP(i)
			{
			/* process withdrawls */
			ANN_WITH_LIST_DEL_WALK(p_rt,i,p_atl,p_rt_next)
				{
				p_rt_next = p_rt->p_with;
				ph1_with_ext_route(p_rt,&out_ann_list);
				changes++;
				}
		
			/* process announces */
			ANN_ANN_NEXT_NLRI(p_rt,i,p_atl)
				{
				if (p_rt->status & IDRP_STATUS_REPLACE)
					{
					/* handle replaces */
					ph1_with_rep_ext_route(p_rt,&out_ann_list);
					}
				else
					{
					/* handle additions */
					/*******/ ph1_add_ext_route(p_rt,&out_ann_list);
					}
				changes++;
				} ANN_ANN_NEXT_NLRI_END;
			} NLRI_FAMILY_LOOP_END 
		} ANN_LIST_END;

		send_best_ext(&out_ann_list);
	
		/* delete the best_external routes which are no 
		 * longer needed after being withdrawn during the 
		 * sending of the new best_ext routes 
		 */
 
		del_routes_best_ext(&out_ann_list);
	
	return(changes);
}


void ph1_with_int_route(p_idrp_rt)
idrpRoute	*p_idrp_rt;
{
idrpPeer	*peer;
#ifndef	ADN_PATCH /* illegal type combination
			&	useless var */
u_int		rtbits;
#endif /* ADN_PATCH */
rt_entry	*p_rt;
int		status_illegal;

	/* peer and rt_entry for gated table */

	peer = p_idrp_rt->peer;
	p_rt = p_idrp_rt->p_rt;
	status_illegal = IDRP_STATUS_EXT_INFO | IDRP_STATUS_EXT_INFO | IDRP_STATUS_MIN_ADV_RUN | IDRP_STATUS_REPLACE;
	if  (p_idrp_rt->status & status_illegal)
		{
		trace (TR_IDRP | TR_UPDATE,0, "idrp ph1_with_int_route - illegal status %x peer %s route_id %d", 
			 p_idrp_rt->status,peer->name,p_idrp_rt->route_id_in); 
		return;
		}
		
	if (!(p_idrp_rt->status & IDRP_STATUS_LOC_RIB)) 
		{
		/* 
		 * Not local rib - so free this idrp route now
		 * It will go away later
		 * clear bits in gated protocol 
		 * free idrpRoute structure
		 */

		idrp_del_rt_gated(p_idrp_rt);	
		idrp_free_nlri_att_rec(p_idrp_rt,NLRI_LINKED_P_WITH); /* free the nlri from the route */
						 /* via withdrawl */
		idrp_free_idrpRoute(p_idrp_rt);	
		}

	else if (p_idrp_rt->status & IDRP_STATUS_LOC_RIB)	
		{
		if (!p_rt)
			{
			trace (TR_IDRP,0,"IDRP(ph1_with_int_route) peer %s has no gated route for nlri %s", 
				peer->name,iso_ptoa(&p_idrp_rt->nlri));

			idrp_free_nlri_att_rec(p_idrp_rt,NLRI_LINKED_P_WITH); /* free the nlri from the route */
							 /* via withdrawl */
			idrp_free_idrpRoute(p_idrp_rt);	
		
			}
		
		/* local rib */
		if (p_rt->rt_state == RTS_DELETE)
			{
			/* put warning flags in all over */
			trace (TR_IDRP | TR_UPDATE, LOG_WARNING, "IDRP (ph1_with_int_route) route from peer %s has route id %d with already delete flag for nlri %s", 
			peer->name,p_idrp_rt->route_id_in,iso_ptoa(&p_idrp_rt->nlri));
			}
 		if (!(p_rt->rt_state & RTS_EXTERIOR)) 
			{
			trace (TR_IDRP|TR_UPDATE, LOG_WARNING, "IDRP (ph1_with_int_route) peer %s does not have exterior flag for nlri %s", peer->name,p_idrp_rt->route_id_in,iso_ptoa(&p_idrp_rt->nlri));
			}

		/* delete this route - it will be flashed if 
		 * the status is local rib 
		 */

		rt_delete(p_rt);
		}
 
	return;
}

/*
 * ph1_with_rep_int_route -- Withdrawal with replace of internal route.
 * 
 * Called from: phase1_internal().
 * 
 */

void
ph1_with_rep_int_route(p_idrp_rt)
idrpRoute	*p_idrp_rt;
{
idrpPeer	*peer;
idrp_rt_chain_walk	*p_gate;
idrp_rt_chain_walk	*p_pref;
sockaddr_un		*p_dest;
idrpRoute		*p_best_ext;
int			pref;

	
	/* phase1 - receives a route with REPLACE status flag set  and
	 *	    p_with with the withdrawn route location
	 *	1) allocate memory for gated and preference chains
	 *	2) try to find the route in the gated tables
	 *	    - it should be there
	 *	    - quit the task if it is not
	 *	3) mediate preference
	 *	4) call idrp_rt_repl_route - replace the
	 *	    route in the gated table for this peer
	 */

	peer = p_idrp_rt->peer;
	p_gate = (idrp_rt_chain_walk *) task_mem_malloc(peer->task,sizeof(idrp_rt_chain_walk));
	p_pref = (idrp_rt_chain_walk *) task_mem_malloc(peer->task,sizeof(idrp_rt_chain_walk));
	p_dest = (sockaddr_un *) task_mem_malloc(peer->task,sizeof(struct iso_net_addr));
	
	if (!idrp_add_route_locate(p_dest,p_idrp_rt,p_gate,p_pref,&p_best_ext))
		{
		/* quit the task - you should find this route in the gated tables */

		trace (TR_IDRP,LOG_WARNING,"IDRP:ph1_with_rep_int_route: peer %s route %s no gated rt", 
			peer->name,iso_ptoa(&p_idrp_rt->nlri));
		task_quit(0);
		}

	if (p_gate->p_rt)
		{	
		trace (TR_IDRP,LOG_WARNING,"IDRP:ph1_with_rep_int_route: peer %s route %s no gated rt for this peer", 
			peer->name,iso_ptoa(&p_idrp_rt->nlri));
		task_quit(0);
		}

	if (p_pref->p_rt)
		{
		pref = idrp_to_gated_pref(p_idrp_rt);
		if (mediate_pref_match(p_pref,p_idrp_rt))
			{
			/* if it is the best preference - set to plus one */
			/* otherwise -  set value to pref minus 1 */
			pref = pref +1;
			}
		else
			{
			pref = pref -1;
			}
			
		free_rt_chain_walk(p_pref->p_next,peer);
		task_mem_free(peer->task,p_pref);
		}

	if (p_gate->p_rt != p_idrp_rt->p_with->p_rt)
		{
		trace (TR_IDRP,LOG_WARNING,"IDRP:ph1_with_rep_int_route: peer %s route %s gated rt doesnt match withdrawl", 
			peer->name,iso_ptoa(&p_idrp_rt->nlri));
		task_quit(0);
		}

	/* get rid of intermediate structures */

	free_rt_chain_walk(p_gate->p_next,peer);
	task_mem_free(peer->task,p_gate);
	task_mem_free(peer->task,p_dest);


	/* replace in the gated route the old idrp route with new idrp route 
	 * - free old idrp route from the attribute record
	 * - free the memory space 
	 */


	idrp_repl_rt_gated(p_idrp_rt,p_idrp_rt->p_with->p_rt);
	idrp_mod_rt_gated(p_idrp_rt,pref);

	/* get rid of the old route */

	idrp_free_nlri_att_rec(p_idrp_rt->p_with,NLRI_LINKED_P_WITH);
	idrp_free_idrpRoute(p_idrp_rt->p_with);	
	trace (TR_IDRP,0,"IDRP:ph1_with_rep_int_route: peer %s route %s gated replaced",peer->name, 
			iso_ptoa(&p_idrp_rt->nlri)); 

}

void 
ph1_add_int_route(p_idrp_rt)
idrpRoute	*p_idrp_rt;
{
idrp_rt_chain_walk	gateway;		/* chain for route locate search for gateway */ 
idrp_rt_chain_walk	preference;		/* chain for route locate serach for perference */
sockaddr_un	*p_dest;			/* destination address for gated Route search */
sockaddr_un	*p_mask;			/* mask for gated table Route search */
rt_entry	*p_rt;				/* gated route entry */
idrpRoute	*p_idrp_rt_old;			/* pointer to old route */
#ifndef ADN_PATCH	/* illegal type combination
			   & not used !!! */ 
rt_parms	rt_parms;			/* rt parms for gated add */
#endif /* ADN_PATCH */
int		pref = 0;				/* preference for gated */
idrpPeer	*peer;				/* peer for route */
idrpRoute	*p_best_rt;			/* pointer to best route for location for idrp */

	
	peer = p_idrp_rt->peer;
	p_dest = (sockaddr_un *) task_mem_malloc(peer->task,ISO_SOCK_SIZE);
	idrp_add_route_locate(p_dest,p_idrp_rt,&gateway,&preference,&p_best_rt);
	pref = idrp_to_gated_pref(p_idrp_rt);
	if (preference.p_rt)
		{
		if (mediate_pref_match(&preference,p_idrp_rt))
			{
			/* if it is the best preference - set to plus one */
			/* otherwise -  set value to pref minus 1 */
			pref = pref +1;
			}
		else
			{
			pref = pref -1;
			}
			
		free_rt_chain_walk(preference.p_next,peer);
		}

	if (gateway.p_rt)
		{
		/* check that it really is a change */

		p_rt = gateway.p_rt;
		if (idrp_rt_change(p_rt,p_idrp_rt))
			{
			/* unlink old idrp route and  link in new route 
			 */
			
			p_idrp_rt_old = idrp_repl_rt_gated(p_idrp_rt, p_rt);
			idrp_free_nlri_att_rec(p_idrp_rt_old,NLRI_LINKED_P_NEXT); /*free attribute linked through p_nlri */ 
			idrp_free_idrpRoute(p_idrp_rt_old);
			idrp_mod_rt_gated(p_idrp_rt,pref); 
			}
		
		else
			{
			trace (TR_IDRP | TR_UPDATE,0, "peer %s route %s said changed but didn't really", 
				peer->name,iso_ptoa(&p_idrp_rt->nlri));
			}
		free_rt_chain_walk(gateway.p_next,peer);
		return;
		} /* end of if p_gate.p_rt - found a gateway match */
	else
		{
		idrp_add_rt_to_gated(p_idrp_rt,p_dest);
		}
	return;

}

/* ph1_with_ext_route
 * Withdraw an external route 
 *
 * (Merit Design Doc Section 4.3.3.2)
 */
void
ph1_with_ext_route(p_idrp_rt,p_ext_ann_list)
idrpRoute	*p_idrp_rt;
idrp_ann_list	*p_ext_ann_list;
{
/* -- skh fixed and refixed - Please look at John
 */
 
idrpPeer	*peer;
#ifndef	ADN_PATCH /* illegal type combination
			&	useless var */
u_int		rtbits;
#endif /* ADN_PATCH */
rt_entry	*p_rt;
int		status_illegal;
idrp_rt_chain_walk	*p_gate;	/* return chain for matched gateways */
idrp_rt_chain_walk	*p_pref;	/* return chain for matched idrp preferences */ 
idrpRoute		*p_best_ext;	/* best external route  */
u_int			status;

	/* peer and rt_entry for gated table */

	peer = p_idrp_rt->peer;
	p_rt = p_idrp_rt->p_rt;
	p_pref = (idrp_rt_chain_walk *) task_mem_malloc(peer->task,sizeof(idrp_rt_chain_walk));
	p_gate = (idrp_rt_chain_walk *) task_mem_malloc(peer->task,sizeof(idrp_rt_chain_walk));

	/* Is this route in AdjRib, Best External and Local Rib? */

	status = p_idrp_rt->status & IDRP_STATUS_INT_MASK; 
	if (status == (IDRP_STATUS_ADJ_RIB | IDRP_STATUS_BEST_EXT | IDRP_STATUS_LOC_RIB))
		{
		/* AdjRib, BestExternal and Local Rib */
 
		if (!idrp_with_route_locate(p_idrp_rt->p_rt->rt_dest, p_idrp_rt, p_gate, p_pref, &p_best_ext))
			{
			/* no route found for this nlri - logic error in code */
			trace (TR_IDRP | TR_UPDATE, LOG_WARNING, "IDRP (ph1_with_ext_route) no route to NLRI %s from peer %s", iso_ptoa(&p_idrp_rt->nlri), peer->name);
			task_quit(0);
			}
		
		if (p_gate->p_rt != p_idrp_rt->p_rt || p_gate->p_next->p_rt)  
			{
			/* 2nd route this destination and this gateway -- it's illegal */
			trace (TR_IDRP | TR_UPDATE, LOG_WARNING, "IDRP (ph1_with_ext_route) duplicate route to NLRI %s from peer %s", iso_ptoa(&p_idrp_rt->nlri), peer->name);
			task_quit(0);
			}

		/* find the next best external for phase1 announcement */
 
		p_best_ext = find_next_best(p_idrp_rt);

		/* set the withdraw flag in this route */
		
		IDRP_STATUS_BIT_SET(p_idrp_rt, IDRP_STATUS_WITH);

		if (p_best_ext)
			{
			/* We have a new best external route 
			 * - 1) flag the new best external,
			 *   2) link new route to phase1 announce lsit
			 *   3) clear this best external 
			 *   4) delete this route - (idrp flash will do final delete )
			 * - watch that this route gets idrp route cleared  
			 */

			IDRP_STATUS_BIT_SET(p_best_ext, IDRP_STATUS_BEST_EXT);
			link_ann_list(p_best_ext, p_ext_ann_list, NLRI_LINKED_P_ANN);
			IDRP_STATUS_BIT_SET(p_idrp_rt, IDRP_STATUS_DELETE);
			idrp_del_rt_gated(p_idrp_rt);
			}
		else
			{
			/* no replacement for this route, it was the only one */
			link_ann_list(p_idrp_rt, p_ext_ann_list, NLRI_LINKED_P_WITH);
			IDRP_STATUS_BIT_SET(p_idrp_rt, IDRP_STATUS_DELETE);
			idrp_del_rt_gated(p_idrp_rt);
			}
		} /* if in Adj_RIB and Best_Ext and Loc_RIB */
	
	else if (status == (IDRP_STATUS_ADJ_RIB | IDRP_STATUS_BEST_EXT))
		{
		/* locate and associated routes to the same destination */

		if (!idrp_with_route_locate (p_idrp_rt->p_rt->rt_dest,p_idrp_rt, p_gate, p_pref, &p_best_ext))
			{
			trace (TR_IDRP | TR_UPDATE, LOG_WARNING, "IDRP (ph1_with_ext_route) no gated rt NLRI %s from peer %s", iso_ptoa(&p_idrp_rt->nlri), peer->name);
			task_quit(0);
			}
			
		if ((p_gate->p_rt != p_idrp_rt->p_rt) && !p_idrp_rt->p_min_adv)
			{
			/* 2nd route this destination and this gateway -- it's illegal unless	*/
			/* min_route_advertisement is set on the route				*/
			trace (TR_IDRP | TR_UPDATE, LOG_WARNING, "IDRP (ph1_with_ext_route) duplicate/no min adv timer route to NLRI %s from peer %s", iso_ptoa(&p_idrp_rt->nlri), peer->name);
			task_quit(0);
			}

		if (p_best_ext != p_idrp_rt)
			{
			trace (TR_IDRP | TR_UPDATE, LOG_WARNING, "IDRP (ph1_with_ext_route) not best external but flag set NLRI %s from peer %s", iso_ptoa(&p_idrp_rt->nlri), peer->name);
			}

		/* find the new best external since this route is deleted */
 
		p_best_ext = find_next_best(p_idrp_rt);


		if (p_best_ext)
			{
			/* We have a new best external route available
			 * 1) flag best external in new route
			 * 2) link to announce list
			 * 3) delete the gated route
			 * 4) free the idrp route structures 
			 */
			IDRP_STATUS_BIT_SET(p_best_ext, IDRP_STATUS_BEST_EXT);
			link_ann_list(p_best_ext, p_ext_ann_list, NLRI_LINKED_P_ANN);
			idrp_del_rt_gated(p_idrp_rt);	
			idrp_free_outlist(p_idrp_rt);	
			idrp_free_nlri_att_rec(p_idrp_rt, NLRI_LINKED_P_WITH);
			idrp_free_idrpRoute(p_idrp_rt);
			}
		else
			{
			/* no replacement for this route, it was the only one */
			IDRP_STATUS_BIT_SET(p_idrp_rt, IDRP_STATUS_WITH);
			IDRP_STATUS_BIT_SET(p_idrp_rt, IDRP_STATUS_DEL_SEND);
			link_ann_list(p_idrp_rt, p_ext_ann_list, NLRI_LINKED_P_WITH);
                        }
		} /* if in Adj_RIB and Best_Ext */

	else if (p_idrp_rt->status & IDRP_STATUS_ADJ_RIB)
		{
			/* only in AdjRib so just delete the route right now */
 
			idrp_del_rt_gated(p_idrp_rt);
			idrp_free_nlri_att_rec(p_idrp_rt, NLRI_LINKED_P_WITH);
			idrp_free_idrpRoute(p_idrp_rt);
		} /* if in Adj_RIB */

	else /* not in Adj_RIB, must be an error */
		{
			trace (TR_IDRP | TR_UPDATE, LOG_WARNING, "IDRP (ph1_with_ext_route) withdrawn route to NLRI %s from peer %s is not in Adj_RIB", iso_ptoa(&p_idrp_rt->nlri), peer->name);
		}

	free_rt_chain_walk(p_pref->p_next,peer);
	free_rt_chain_walk(p_gate->p_next,peer);
	task_mem_free(peer->task,p_pref);
	task_mem_free(peer->task,p_gate);	
}


/*
 * ph1_with_rep_ext_route -- withdraw with replace of external route.
 *
 * Called from: phase1_external().
 *
 * (Merit IDRP Design Doc 4.3.2.2)
 */

void 
ph1_with_rep_ext_route(p_idrp_rt,p_ext_ann_list)
idrpRoute 	*p_idrp_rt;
idrp_ann_list	*p_ext_ann_list;
{
idrpPeer	*peer;
#ifndef ADN_PATCH /* illegal type combination
                        &       useless var */
u_int		rtbits;
#endif /* ADN_PATCH*/
rt_entry	*p_rt;
int		status_illegal;
idrp_rt_chain_walk	*p_gate;	/* return chain for matched gateways */
idrp_rt_chain_walk	*p_pref;	/* return chain for matched idrp preferences */ 
sockaddr_un		*p_dest;
idrpRoute		*p_best_ext;	/* best external route  */
int			pref;


	peer = p_idrp_rt->peer;
	p_rt = p_idrp_rt->p_rt;

	/* initialization of routine lists */
	p_gate = (idrp_rt_chain_walk *) task_mem_malloc(peer->task,sizeof(idrp_rt_chain_walk));
	p_pref = (idrp_rt_chain_walk *) task_mem_malloc(peer->task,sizeof(idrp_rt_chain_walk));
	p_dest = (sockaddr_un	*) task_mem_malloc(peer->task,sizeof(struct iso_net_addr));	

	if (!idrp_add_route_locate(p_dest,p_idrp_rt,p_gate,p_pref,&p_best_ext))
		{
		/* quit the task - you should find this route in the gated tables */

		trace (TR_IDRP,LOG_WARNING,"IDRP:ph1_with_rep_ext_route: peer %s route %s no gated rt", 
			peer->name,iso_ptoa(&p_idrp_rt->nlri));
		task_quit(0);
		}
	if (p_gate->p_rt != p_idrp_rt->p_rt)
		{	
		trace (TR_IDRP,LOG_WARNING,"IDRP:ph1_with_rep_int_route: peer %s route %s no gated rt for this peer", 
			peer->name,iso_ptoa(&p_idrp_rt->nlri));
		task_quit(0);
		}
	pref = idrp_to_gated_pref(p_idrp_rt);
	if (p_pref->p_rt)
		{
		if (mediate_pref_match(p_pref,p_idrp_rt))
			{
			/* if it is the best preference - set to plus one */
			/* otherwise -  set value to pref minus 1 */
			pref = pref +1;
			}
		else
			{
			pref = pref -1;
			}
		}

	if (p_gate->p_rt != p_idrp_rt->p_with->p_rt)
		{	
		trace (TR_IDRP,LOG_WARNING,"IDRP:ph1_with_rep_ext_route: peer %s route %s gated rt doesnt match withdrawn route", 
			peer->name,iso_ptoa(&p_idrp_rt->nlri));
		task_quit(0);
		}

	free_rt_chain_walk(p_pref->p_next,peer);	
	free_rt_chain_walk(p_gate->p_next,peer);	
	task_mem_free(peer->task,p_dest);
	task_mem_free(peer->task,p_pref);
	task_mem_free(peer->task,p_gate);
	idrp_replace_ext(p_idrp_rt,p_idrp_rt->p_with->p_rt,p_ext_ann_list,p_best_ext,pref);
}


/*
 * ph1_add_ext_route -- phase1 add external route
 *
 * Called from: phase1_external()
 */

void ph1_add_ext_route(p_idrp_rt,p_ext_ann_list)
idrpRoute 	*p_idrp_rt;
idrp_ann_list	*p_ext_ann_list;
{
idrp_rt_chain_walk	gateway;		/* chain for route locate search for gateway */ 
idrp_rt_chain_walk	preference;		/* chain for route locate search for preference */
struct iso_net_addr	dest;			/* destination address for gated Route search */
rt_entry	*p_rt;				/* gated route entry */
idrpRoute	*p_idrp_rt_old;			/* pointer to old route */
#ifndef ADN_PATCH       /* illegal type combination
                           & not used !!! */ 
rt_parms	rt_parms;			/* rt parms for gated add */
#endif /* ADN_PATCH */
int		pref;				/* preference for gated */
idrpRoute	*p_best_rt;			/* pointer to best route for location for idrp */
idrpPeer	*peer;

	
	peer = p_idrp_rt->peer;
	trace (TR_IDRP,0,"Called ph1_with_add_ext_route peer %s route_id %d", 
			peer->name,p_idrp_rt->route_id_in);

	/* set AdjRib flag at beginning */

	idrp_add_route_locate((sockaddr_un *) &dest,p_idrp_rt,&gateway,&preference,&p_best_rt);

	/* If existing routes with the same preference were found, then do a tie-break. */

	pref = idrp_to_gated_pref(p_idrp_rt);
	if (preference.p_rt)
		{
		if (mediate_pref_match(&preference,p_idrp_rt))
			{
			/* if it is the best preference - set to plus one 
			 * otherwise -  set value to pref minus 1 
			 */
			pref = pref +1;
			}
		else
			{
			pref = pref -1;
			}
			
		free_rt_chain_walk(preference.p_next,peer);
		}

	/* If the "new" route is a replacement,
	 *  that is from a gateway  that has already given us this rt
	 */

	if (gateway.p_rt)
		{
		/* Implicit Replace 
		 * - Check to see if we really need to
		 *   change this route
		 * 	(idrp_rt_change - checks the idrp_attributes
		 * 	to see if this is really a change)
		 */

		p_rt = gateway.p_rt;
		if (idrp_rt_change(p_rt,p_idrp_rt))
			{
			idrp_replace_ext(p_idrp_rt,p_rt,p_ext_ann_list,p_best_rt,pref);
			}
		else
			{
			trace (TR_IDRP | TR_UPDATE,0, "IDRP peer %s route %s said changed but didn't really", 
				peer->name,iso_ptoa(&p_idrp_rt->nlri));
			}


		free_rt_chain_walk(gateway.p_next,peer);
		} /* end of if p_gate.p_rt - found a gateway match */
	else
		{
		/* True add of external route to gated  */ 
		 

		/*  Do other external routes exist?
		 */

		if (!p_best_rt)	
			{
			/* no best external should mean no external route */
			
			if (gateway.p_next)
				{
				trace(TR_IDRP,0, "Best external destination %s peer %s broken no route added ",
					iso_ptoa(&p_idrp_rt->nlri),peer->name);
				return;
				}
			
			/* If we have reached this point this is the only route to this NLRI, so it is by 
			 * definition the best external route.  Create the idrp_best_ext struct and
			 * set this route to be the best ext.  Link it into the BER-Rib. */

			/* Create the idrp_best_ext struct and link the route into the chain */
			p_idrp_rt->p_p_best_ext = (idrp_best_ext *) task_mem_malloc(idrp_this_node.task, sizeof(idrp_best_ext));
			p_idrp_rt->p_p_best_ext->p_best_ext = p_idrp_rt;	/* this must be the best ext */
			p_idrp_rt->p_p_best_ext->refcount = 1;			/* only we refer to it so far */
			p_idrp_rt->p_better = p_idrp_rt->p_next_best = NULL;	/* no other routes on the chain */

			/* Now link into BER-Rib, unsorted */
			p_idrp_rt->p_p_best_ext->p_next = ber_rib_head; 
			p_idrp_rt->p_p_best_ext->p_prev = NULL;
			if (ber_rib_head != NULL)
				{
				ber_rib_head->p_prev = p_idrp_rt->p_p_best_ext;
				}
			ber_rib_head = p_idrp_rt->p_p_best_ext;

			IDRP_STATUS_BIT_SET(p_idrp_rt,IDRP_STATUS_BEST_EXT);
			link_ann_list(p_idrp_rt,p_ext_ann_list,NLRI_LINKED_P_ANN);
			}
		else
			{
			/* Best external route exists 
			 * - see if this route is better
			 * - if so, then reset best_external for destination
			 *       and link to announce list
			 * - if not, just add to gated table
			 */

			if (idrp_pref_compare(p_idrp_rt,p_best_rt))
				{
				/* new route is the best external route */
				/* If we have reached this point this is the best of a set of routes
				 * to this NLRI.  Link it into the list and point the BER pointer to it */
				/* NOTE:  I'm assuming below that p_best_rt points to the best EXISTING 
				 * route.  Check this! */
				p_idrp_rt->p_next_best = p_best_rt;	/* put this one at the top of the best routes list */
				p_best_rt->p_better = p_idrp_rt;	/* previous best now is second in line */
				p_idrp_rt->p_better = NULL;		/* no one is better than this one */
				p_idrp_rt->p_p_best_ext = p_best_rt->p_p_best_ext;	/* find pre-existing idrp_best_ext struct */
				p_idrp_rt->p_p_best_ext->p_best_ext = p_idrp_rt;	/* and point it to this route */

				IDRP_STATUS_BIT_CLEAR(p_best_rt,IDRP_STATUS_BEST_EXT);
				IDRP_STATUS_BIT_SET(p_idrp_rt,IDRP_STATUS_BEST_EXT);
				link_ann_list(p_idrp_rt,p_ext_ann_list,NLRI_LINKED_P_ANN);
				}
			else
				{
				/* new route is not best external route, but we still need to find its place in the sun */
				/* I am assuming that the prefs have been correctly set on the routes already */
				insert_in_pref_order(p_idrp_rt, p_best_rt);
				}
			}
			
		idrp_add_rt_to_gated(p_idrp_rt,(sockaddr_un *) &dest);
		}


	return;

}


/* idrp_replace external routes 
 * - replace routes for implict or explicit
 *   replace of external routes
 *    
 * (Merit IDRP Design Doc 4.3.3.2)
 */

void 
idrp_replace_ext(p_idrp_rt,p_rt,p_ext_ann_list,p_best_rt,pref)
idrpRoute	*p_idrp_rt;	/* idrpRoute replacing old route */
rt_entry	*p_rt;		/* gated rt_entry for old idrpRoute being replaced */
idrp_ann_list	*p_ext_ann_list;/* idrp ann list for ph 1 ext routes to int peers */
idrpRoute	*p_best_rt;	/* current bext ext route */
int		pref;		/* mediated preference */
{
idrpRoute	*p_old_idrp_rt;	/* old idrp route being replaced */
idrpRoute	*p_best;		/* scratch to hold new best route */
int		status;

	p_old_idrp_rt = (idrpRoute*) p_rt->rt_data;
	status = p_old_idrp_rt->status & IDRP_STATUS_INT_MASK;
	trace (TR_IDRP,0,"idrp_replace_ext peer %s nlri %s new pref %d status %d", 
			p_idrp_rt->peer->name,iso_ptoa(&p_idrp_rt->nlri),pref,status);

	/* handle the route based on the old route status */
  
	switch (status)
		{
		case (IDRP_STATUS_ADJ_RIB):
			{
			if (idrp_pref_compare(p_idrp_rt, p_best_rt))
				{
				/* This route is the (new) best external */
				IDRP_STATUS_BIT_SET(p_idrp_rt, IDRP_STATUS_BEST_EXT);
				IDRP_STATUS_BIT_CLEAR(p_old_idrp_rt, IDRP_STATUS_BEST_EXT);
							
				/* Relink the sorted-by-pref chain for this NLRI     */
				/* %jgs% -- refcount doesn't change since this is a  */
				/* replace?  or does it?  depends on what/where/when */
				/* /how we free the replaced route 		     */
				p_best->p_better = p_idrp_rt;
				p_idrp_rt->p_next_best = p_best;
				p_idrp_rt->p_better = NULL;
				p_idrp_rt->p_p_best_ext = p_best->p_p_best_ext;
				p_idrp_rt->p_p_best_ext->p_best_ext = p_idrp_rt;
			
				link_ann_list(p_idrp_rt, p_ext_ann_list, NLRI_LINKED_P_ANN);
				} /* if pref_compare */
			break;
			}

		case (IDRP_STATUS_ADJ_RIB | IDRP_STATUS_BEST_EXT | IDRP_STATUS_LOC_RIB):
		case (IDRP_STATUS_ADJ_RIB | IDRP_STATUS_BEST_EXT):
			{

			if ((p_old_idrp_rt->p_next_best == NULL) ||
			 (idrp_pref_compare(p_idrp_rt, p_old_idrp_rt->p_next_best)))
				{
				/* this is (still) the best external route */
				IDRP_STATUS_BIT_SET(p_idrp_rt, IDRP_STATUS_BEST_EXT);
				
				/* make sure that best external pointer still points at this	*/
				/* replacement route.  Link it into the sorted-by-pref chain.	*/
				p_idrp_rt->p_p_best_ext = p_old_idrp_rt->p_p_best_ext;
				p_idrp_rt->p_p_best_ext->p_best_ext = p_idrp_rt;
				p_idrp_rt->p_next_best = p_old_idrp_rt->p_next_best;
				p_idrp_rt->p_better = p_old_idrp_rt->p_better;
				if (p_idrp_rt->p_next_best)
					{
					p_idrp_rt->p_next_best->p_better = p_idrp_rt;
					}
							
				link_ann_list(p_idrp_rt, p_ext_ann_list, NLRI_LINKED_P_ANN);
				}
			else
				{
				/* the new route is no longer the best external */
				p_best = find_next_best(p_old_idrp_rt);
				
				IDRP_STATUS_BIT_SET(p_best, IDRP_STATUS_BEST_EXT);
				
				/* Update p_best in the sorted-by-pref chain.		*/
				p_best->p_p_best_ext->p_best_ext = p_idrp_rt;
				p_best->p_better = NULL;
	
				/* Insert p_idrp_rt into the sorted-by-pref chain	*/
				insert_in_pref_order(p_idrp_rt, p_best);
				
				link_ann_list(p_best, p_ext_ann_list, NLRI_LINKED_P_ANN);
				}
			break;
			}

		}
	
	/* replace the idrp route in gated route */
	idrp_repl_rt_gated(p_idrp_rt, p_rt);
	
	/* change gated route so we flash if gated route changes */
	idrp_mod_rt_gated(p_idrp_rt,pref);	

	/* free up the old idrp route */

	idrp_free_nlri_att_rec(p_old_idrp_rt,NLRI_LINKED_P_WITH);
	idrp_free_idrpRoute(p_old_idrp_rt);
}


/*
 * send_best_ext(idrp_ann_list *p_ext_ann_list)
 *
 * Send best external routes to internal IDRP neighbors.
 *
 * (Merit IDRP design document 4.3.4.1, documented as "idrp_send_phase1_ann")
 */
void 
send_best_ext(p_ext_ann_list)
idrp_ann_list	*p_ext_ann_list;
{
idrpPeer	*peer;
idrp_ann_list	*p_atl;
idrpRoute	*p_wdraw;
idrpRoute	*p_irt;
int		family;
idrp_send_list	*p_send_list = NULL;

	/* send announce list of best_ext routes to each idrp internal peer */
	
	trace (TR_IDRP,0,"send_best_ext called ");

	IDRP_INT_PEER(peer)
		{
		ANN_LIST(p_atl,p_ext_ann_list)
			{
			NLRI_FAMILY_LOOP(family)
				{
				ANN_WITH_LIST_WALK(p_wdraw,family,p_atl)
					{
						/* for each withdraw NLRI/idrpRoute structure, 
						 * process withdraw into send list for a single
						 * update bispdu.  If update bispdu is full, it
						 * will be set to neighbor (send_with_attr).
						 */
						 p_send_list = send_with_attr(p_wdraw, p_atl->p_attr, p_send_list,peer);
					} ANN_WITH_LIST_WALK_END;

				ANN_ANN_LIST_WALK(p_irt,family,p_atl)
					{
					/* process each nlri/idrproute structure on announce
					 * list and add to a send list for a single update
					 * bispdu.  If the update bispdu is full, it will
					 * be sent to neighbor (send_nlri_attr).
					 */
					 p_send_list = send_nlri_attr(p_irt,p_atl->p_attr,p_send_list,peer);
					} ANN_ANN_LIST_WALK_END; 

				} NLRI_FAMILY_LOOP_END; 
			/* Flush the send list before starting on a new attribute record.
			 * Any outstanding announcements will be stuffed into a PDU and sent.
			 */

			flush_att_send_list(peer,p_atl->p_attr, p_send_list);

			} ANN_LIST_END; 
		} IDRP_INT_PEER_END;
}
	 

/*
 * del_routes_best_ext(idrp_ann_list *p_ext_ann_list)
 *
 * Delete defunct best external routes.  We only need to do this for BERs that
 * were not in the LOC_Rib when they were withdrawn, so gated will not flash
 * us to inform us that they were deleted.
 *
 * (Merit IDRP design document 4.3.4)
 */
void del_routes_best_ext(p_ext_ann_list)
idrp_ann_list	*p_ext_ann_list;
{
idrp_ann_list	*p_atl;
idrpRoute	*p_wdraw,*p_rt_next;
int		family;

	/* delete routes on best external routes to each idrp internal peer */

	trace (TR_IDRP,0,"del_routes_best_ext called ");

	/* "Walk withdraw route NLRI list doing... */

	ANN_LIST(p_atl,p_ext_ann_list)
		{
		NLRI_FAMILY_LOOP(family)
			{
			ANN_WITH_LIST_DEL_WALK(p_wdraw,family,p_atl,p_rt_next)
				{
					/* --- skh-- why use this coding construct??
					 */
					p_rt_next = p_wdraw->p_with;

					if (p_wdraw->status & IDRP_STATUS_DEL_SEND)
						{
						idrp_del_rt_gated(p_wdraw);
						idrp_free_nlri_att_rec(p_wdraw, NLRI_LINKED_P_WITH);
						idrp_free_outlist(p_wdraw);
						idrp_free_idrpRoute(p_wdraw);
						}

				} ANN_WITH_LIST_WALK_END ;
			} NLRI_FAMILY_LOOP_END;
		} ANN_LIST_END; 
}

idrpRoute *
remove_ann_dup(p_withdraw,p_announce)
idrpRoute *p_withdraw;
idrpRoute *p_announce;
{
idrpRoute	*p_w_rt;
idrpRoute	*p_w_rt_last = 0;
idrpRoute	*p_a_rt;
idrpRoute	*p_with_return = 0;
int		found;

	p_w_rt = p_withdraw;
	p_a_rt = p_announce;

	/* walk withdrawal list checking for announce nlris */
	/* if match found, link change entry in announce list with */
	/* 		   flag the withdraw/replace in status */ 
	/* if no match found, link withdraw to withdrawl list */

	while (p_w_rt)
		{
		/* loop while checking announce nlris */
		found = FALSE;
		while (p_a_rt)
			{
			if (p_w_rt == p_a_rt)
				{
				/* match found - set status of announce route */ 
				/* save the withdraw route id in announce record */
				found = TRUE;
				p_a_rt->status = IDRP_STATUS_REPLACE;
				p_a_rt->p_with = p_w_rt;
				p_a_rt = 0;
				}
			else
				{
				/* no match found keep search announces */
				p_a_rt->p_next_nlri;
				}
			}

		if (!found)
			{
			/* not found, add this id to withdraw list */

			if (p_with_return)
				{
				/* list exists - so link */ 
				/* clear the route_id in link */
				p_with_return->p_with = p_w_rt;
				}
			else
				{
				/* first entry in list so init */
				p_with_return = p_w_rt;
				}

			/* clear the withdrawn nlri link */
			/* unlink from route_id next list */
			
			p_w_rt->p_with = (idrpRoute *) 0;
			} /* end of while to check announce list */
		
		p_w_rt_last = p_w_rt;
		p_w_rt = p_w_rt->p_next_nlri;
		p_w_rt_last->p_next_nlri = (idrpRoute *)0;
				
		} /* end of while to walk nlri list */ 

	return (p_with_return);
}		
