
/* idrp_parse_init.c
 * 
 * routines that handle the parsing and configuration
 * set-up for idrp  
 *
 *  John Scudder, Susan Hares Merit/NSFNET. 
 *
 * idrp_find_peer - find the peer structure
 * idrp_link_peer  - link the peer structure
 * idrp_peer_alloc - allocated peer structure out of allocated memory (not with a task)
 * idrp_small_blk_alloc - allocate small block out of allocated memory (not with a task)
 * idrp_med_blk_alloc - allocate medium block out of allocated memory (not with a task)
 * idrp_small_blk_free - free small block of allocated memory (not within a task)
 * idrp_med_blk_fred - free small block of allocated memory (not within a task)
 * idrp_big_blk_free - free peer size block of allocated memory (not within a task)
 * idrp_local_peer_init - initialize the this nodes (local) peer during configuration
 * idrp_delete - delete the peer structure  and relink peer list
 *		      other routines have already turned peer down logically
 * idrp_peer_config_init(peer) - initialize peer configuration values to preset limits 
 * idrp_peer_update(old_peer,peer) - update peer structure from new peer  
 */

#include "include.h"
#include "idrp.h"
#include "inet.h"



/* find the idrp peer among the
 * idrp peer list
 */

idrpPeer *idrp_find_peer(p_new_peer)
idrpPeer	*p_new_peer;
{
idrpPeer	*peer;
	
	IDRP_LIST(peer)
		{
		/* if new peer has iso address - look for duplicate ISO address */

		if (p_new_peer->neighbor.isoa_len)
			{
			if (!isoaddrcmp(&peer->neighbor,&p_new_peer->neighbor))
				{
				/* ISO address compare - check RDIs */

				if (!isoaddrcmp(&peer->rdi,&p_new_peer->rdi))
					{
					/* NET and RDI match -  call matched */ 
					return (peer);
					}
				}
			}
		
		/* if new peer has ip address - look for duplicate ip address */
	
		if (p_new_peer->ip_neighbor)
			{
			if (!sockaddrcmp2(peer->ip_neighbor,p_new_peer->ip_neighbor))
				{
				/* ip sockaddr address match - try RDI */ 
				if (!isoaddrcmp(&peer->rdi,&p_new_peer->rdi))
					{
					return(peer);
					}
				}
			}
		} IDRP_LIST_END;

	/* no match for peer */
	peer = (idrpPeer *) 0;
	return(peer);
}


/* link the new idrp peer to the idrp peer
 * list.  List has order of internal peers and
 * test peers first, and then external peers at
 * end. The internal peer peer list is walked 
 * each time we send external changes to
 * internal neighbors.
 */


void
idrp_link_peer(p_peer)
idrpPeer	*p_peer;
{
idrpPeer	*peer;
idrpPeer	*p_last;
idrpPeer	*p_last_int;

	if (!idrp_peers)
		{
		/* nothing on the peer list -initialize it
		 *
		 */

		idrp_peers = p_peer;
		idrp_neighbors = 1;
		idrp_n_established = 0;
		idrp_n_disabled = 0;
		if (p_peer->type == IDRP_PEER_EXTERNAL)
			{
			idrp_ext_peers = p_peer;
			idrp_ext_neighbors = 1;
			}
		else
			{
			idrp_ext_peers = (idrpPeer *) 0;
			idrp_ext_neighbors = 0;
			}
		p_peer->id = idrp_neighbors;
		return;
		}
			
	IDRP_LIST(peer)
		{
		p_last = peer;
		if (peer->type == IDRP_PEER_INTERNAL)
			{
			p_last_int = peer;
			}
		} IDRP_LIST_END;

	switch (p_peer->type)
		{
		case IDRP_PEER_INTERNAL:
		case IDRP_PEER_TEST:
			p_peer->next = p_last_int->next;
			p_last_int->next = p_peer;
			idrp_neighbors++;
			break;
	
		case IDRP_PEER_EXTERNAL:
			p_last->next = p_peer;
			p_peer->next = (idrpPeer *) 0;
			if (!idrp_ext_peers)
				{
				idrp_ext_peers = p_peer;
				}
			idrp_ext_neighbors++;
			idrp_neighbors++;
			break;
		}
	p_peer->id = idrp_neighbors;
}
	
				
idrpPeer *idrp_peer_alloc(proto)
idrpPeer *proto;
{
	idrpPeer *peer;
	
	if (!idrp_big_block_index) {
		idrp_big_block_index = task_block_init(IDRP_BIG_BLOCK_SIZE);
	}


	if (proto) 
		{
		*peer = *proto;
		}
	else
		{
		peer = (idrpPeer *) task_block_alloc(idrp_big_block_index);
		}
	bzero(peer,sizeof(idrpPeer));

return (peer);
}

/* allocate a small idrp task block structure 
 * (alias baby bear idrp allocate)
 * These blocks are created by the parsing routines
 * in the configuration to store local route information
 * that does not have a task to look after it.
 * This is the small size used for:
 *  - idrpRoutes for local routes
 *  - rd_path structure
 *  - route_list_entry structure   
 */

u_char *idrp_small_blk_alloc()
{
u_char	*p_baby_bear;
	if (!idrp_small_blk_index) {
		idrp_small_blk_index = task_block_init(IDRP_SMALL_BLOCK_SIZE);
		}

	p_baby_bear = task_block_alloc(idrp_small_blk_index);

return(p_baby_bear);
}

/* free a small idrp task block structure 
 *
 * - free the small block
 */
void idrp_small_blk_free(p)
u_char *p;
{

	task_block_free(idrp_small_blk_index,p);
}


/* allocate a medium size task block 
 * (alias mama bear)
 * 
 * This blocks are medium size used by parsing routines 
 * - only attribute block uses this size right now
 */
 
u_char *idrp_med_blk_alloc()
{
u_char	*p_mama_bear;

	if (!idrp_med_blk_index) {
		idrp_med_blk_index = task_block_init(IDRP_MEDIUM_BLOCK_SIZE);
		}

	p_mama_bear = (u_char *) task_block_alloc(idrp_med_blk_index);

return(p_mama_bear);
}	
			
/* free a medium idrp task block structure 
 *
 * - free the medium block
 */
void idrp_med_blk_free(p)
u_char *p;
{

	task_block_free(idrp_med_blk_index,p); 
}


/* free the big block set for peer 
 *
 */
 

void idrp_big_blk_free(peer)
idrpPeer	*peer;
{
	task_block_free(idrp_big_block_index,peer);
}



void idrp_local_peer_init()
{
static char local_name [] = "local_node";
u_short hold;
int	len;
idrp_attribute_record 	*att;
idrpPeer		*peer;

	/* set-up the peer structure */

	/* hack for now - until I get parser working  */ 
	idrp_this_node.nlri_supp[IDRP_NLRI_IP] = TRUE;
	idrp_this_node.nlri_supp[IDRP_NLRI_ISO] = TRUE;

	/* on to rest of configuration */
 
	idrp_this_node.next = (idrpPeer *) 0;
	idrp_this_node.type = IDRP_PEER_LOCAL;
	bcopy(&local_name,&idrp_this_node.name,sizeof(local_name));

	/* set-up the gateway entry for this peer */
	/* sockaddr for interface will need to wait until idrp_init time */

	idrp_this_node.gw.gw_proto =  RTPROTO_IDRP;
	idrp_this_node.iso_gw.gw_proto = RTPROTO_IDRP;
	idrp_this_node.iso_gw.gw_addr = (sockaddr_un *) &idrp_this_node.iso_intf_addr;

	/* this needs to change to be configured into the box */

	idrp_this_node.proto_sock = IDRP_PDU_PROTO_IP_RAW;
	IDRP_LIST(peer)
		{
		if (peer->proto_sock != IDRP_PDU_PROTO_IP_RAW) 
			{
			idrp_this_node.proto_sock = peer->proto_sock;
			break;
			}	
		} IDRP_LIST_END;
	idrp_this_node.iso_intf_addr.isoa_len = rt_net.isoa_len + 2;
	idrp_this_node.iso_intf_addr.isoa_family = AF_ISO;
	bcopy(&rt_net.isoa_genaddr,idrp_this_node.iso_intf_addr.isoa_genaddr,rt_net.isoa_len);

       /* set-up local rd_path for local routes
	* total segement = segement path header length (3 bytes)
	*      + rdi length + 1 byte for RDI length 
	* - segement path = length + length of RDI 
 	* - segment type - sequence
	*/

        len = rt_rdi.isoa_len + (sizeof(idrp_path_segment_header)) + RDI_LENGTH_BYTE_LEN;

	/* here's the local Segment path length - RDI plus one length byte
	 */
 
	local_rd_path.rd_header.path_seg_type = IDRP_PATH_SEGMENT_SEQ;
	hold = rt_rdi.isoa_len + RDI_LENGTH_BYTE_LEN;	
	hold = htons(hold);
       	bcopy(&hold,&local_rd_path.rd_header.path_seg_len,sizeof(local_rd_path.rd_header.path_seg_len));
	trace (TR_PARSE,0,"local_rd_path length = %d  = %x",hold,local_rd_path.rd_header.path_seg_len);
	bcopy(rt_rdi.isoa_genaddr,local_rd_path.rdi.isoa_genaddr,rt_rdi.isoa_len);
	local_rd_path.rdi.isoa_len = rt_rdi.isoa_len;

	/* we set everything to into attribute except the length for
	 * the local rd path without knowing the RDI.  So 
	 * set it now that the parsing has finished
	 */
 
	ATTR_LIST(att)
		{
		if (att->local_mask & IDRP_LOCAL_RDPATH_RESET_LENGTH)
			{
			att->attrib[IDRP_ATTR_RD_PATH].length = len;
			}
		} ATTR_LIST_END;

	/* one more local attribute thing - set the local hop count to zero */
	local_hop_count = 0;	
}     

/*
 *      Delete a peer
 */

void idrp_delete(peer)
idrpPeer *peer;
{
idrpPeer *peer2;
idrpPeer *peer_last;

    /* move relink idrp_peers chain */
 
    if (peer == idrp_peers) {
        idrp_peers = peer->next;
    } else {
 
        IDRP_LIST(peer2) {
            if (peer2->next == peer) {
                peer2->next = peer->next;
                break;
            }
        } IDRP_LIST_END
    }

    /* take care of external neighbor list and count */	
    if (peer->type == IDRP_PEER_EXTERNAL)
	{
	idrp_ext_neighbors--;
	if (idrp_ext_peers == peer)
		{
		idrp_ext_peers = peer->next;
		}		
				
	}
    idrp_neighbors--;
    task_mem_free((task *) 0, (caddr_t) peer);
}

void
idrp_peer_config_init(peer)
idrpPeer *peer;
{

	/* configuration default setting on peer
	 * parsing sets up - peer list, type of peer, address of neighbor,
	 */ 

	peer->version = 1;
	peer->AuthType = 0;
	peer->nlri_supp[0] = AF_ISO;	/* osi support */
	peer->nlri_supp[1] = AF_INET;	/* ip support */

	/* set up the gated gateways so that we can configure the
	 * tasks
	 * - set up an IP gateway and an ISO based gateway
	 * so I can get the deletes through gated
	 */

	peer->gw.gw_proto = peer->iso_gw.gw_proto = RTPROTO_IDRP;
	if (peer->proto_sock == IDRP_PDU_PROTO_IP_RAW)
		{
		peer->gw.gw_addr->in.sin_port = 0;
		}
	else
		{
		peer->gw.gw_addr->in.sin_port = IDRP_PORT;
		}
	/* temporary hack until we get neighbor aligned */

	peer->iso_gw_addr = peer->neighbor;
	peer->iso_gw_addr.isoa_len += 2; 
	peer->iso_gw.gw_addr = (sockaddr_un *) &peer->iso_intf_addr;
	peer->pdu_maxrcvsize = IDRP_PKTSIZE;
	peer->trace_flags = TRUE;
	peer->listen_for_open = TRUE;
}

void idrp_peer_update(peer_old,peer)
idrpPeer	*peer_old;
idrpPeer	*peer;
{

	/* copy in the current peer NET and status from
	 * the new structure to the old
	 * - copy in the whole structure of the iso NET 
	 * - copy in the pointer to the IP address
	 * - copy in the whole RDI structure
	 */
 
	
	bcopy(&peer->neighbor,&peer_old->neighbor,sizeof(struct iso_net_addr));
	peer_old->ip_neighbor = peer->ip_neighbor;
	bcopy(&peer->rdi,&peer_old->rdi,sizeof(struct iso_net_addr));


	/* The flags and protocol socket need special handlign
	 *  1) status of neighbor (idrp_flags)
	 *	Unconfigured: should be illegal
	 *
	 * 	DELETED  - set in idrp_cleanup,
	 *		use the new flags
	 *		re-use peer session
	 *	TRY_CONNECT - set in parser.y 
	 *		use the new flags
	 * 	CONNECT - peer is still connected??
	 *		  set as connected to remote peer before re-config
	 *		  try to keep it going if
	 *		  Or in the new flags 
	 *		  
	 *	WRITEFAILED - peer failed on socket write 
	 *		Use the new flags
	 *
	 *	IDLE - set to disable peer in last config 
	 *		use new flags
	 *
	 * if connect flag set, or in the protocol socket
	 */ 
 
	if (peer_old->idrp_flags & IDRPF_CONNECT)
		{
		/* or in the connect flags */
		peer_old->idrp_flags = IDRPF_CONNECT | peer->idrp_flags;
		idrp_n_established++;
		}
	else
		{
		/* if not connected, just use new flags */
	 
		peer_old->idrp_flags = peer->idrp_flags;
		}
	/* copy last protocol socket open to old_proto_sock 
	 * get new configured one in proto_sock 
	 */ 

	peer_old->last_proto_sock = peer_old->proto_sock;
	peer_old->proto_sock = peer->proto_sock;
}
