#include "include.h"
#include "idrp.h"

/* idrp_rt_ext_info.c
 * - routines that handle  IDRP EXT_INFO processing
 * things includes:
 *  	1) IS-IS <-> IDRP
 *	2) static <-> IDRP
 *	3) interface <-> IDRP
 * 	4) BGP <-> IDRP
 * 
 * In the future, the BGP code will be pulled into
 * a separate file.
 *
 * %% finding the general route 
 * 
 *
 *  
 * %% ext_info routines for the attribute record (gated related)
 *
 * find_ext_info_attr_rec - find the external info attribute record to
 *		   match this one
 *
 * create_ext_info_attr_rec - create external info attribute record
 * 
 * idrp_aspath_to_canon_rdpath - as_path structure to RDI canonical pathway
 *  generic_ext_info(p_can_list,p_att,peer,size) - set-up generic ext_info attribute
 *  idrp_bgp_within_RD(p_idrp_rt) - is this bgp route within this RD
 */



/* %% external information attribute record */

idrp_attribute_record *
find_ext_info_attr_rec(p_rth,p_rt)
rt_head		*p_rth;
rt_entry	*p_rt;
{
idrp_attribute_record 	*p_att;
int			i;
int			match;

	/* Types of EXT_INFO attribute are in the attribute list:
 	 * - 1) AS path of zero - therefore is linked 
	 *   to a local RD EXT_INFO path
	 * - 2) non-zero AS path
	 *       one comes from BGP-3 or BGP-4 with
	 * 	 full AS path or  EGP with single AS path
	 *       OSPF or other might have some internal stuff
	 *	 but it should be a full aspath
	 * 
	 * If you are case one, you can pick up the EXT_INFO
	 * pathway already created for you in the local_peer_init code
	 *   
	 * walk the list looking for an EXT_INFO IDRP attribute 
	 * with local attributes matching the protocol information 
	 * came from
	 */

	if (!p_rt->rt_aspath)
		{
		return(find_local_ext_attr(p_rt->rt_routers[0],(idrpRoute_options *)0));
		}
	else
		{
		if (p_rt->rt_aspath->path_len == 0)
			{
			return (find_local_ext_attr(p_rt->rt_routers[0],(idrpRoute_options *)0));
			}
		}

	/* only AS pathways of one or more get here 
	 *  we have an aspath structure - with ASs 
 	 * so go walking looking for an  EXT_INFO pathway
	 * structure in  in the attribute list
	 */

	ATTR_LIST(p_att)
		{
		if (p_att->idrp_mask & IDRP_ATTBT_EXT_INFO)
			{
			/* we've got an external attribute list
			 * Is it local? 
		 	 */

			if (!(p_att->local_mask & IDRP_LOCAL_EXT_INFO))
				{
				/* not a local EXT_INFO
				 * - try check AS path - pointers are they the same?
				 */

				if (p_att->as_path == p_rt->rt_aspath)
					{
					/* match on AS paths and therefore RD_PATHS
					 * Does it match on the NEXT_HOP?
					 * rtp_routers - is the next hop in gated
					 *  
					 * - sockcompare on next hop for routers
					 * - to make sure res
					 *   Do it in gated land
					 */    

				
					if (sockaddrcmp(p_rt->rt_routers[0],p_att->next_hop))
						{
						/* sockets compare - so return this attribute record
						 * - skh 3/11/93 - When IS-IS QOS comes we need
						 *   to put the QOS test here for the route
						 */

						return (p_att);
						}
					} /* end if as path match */
				} /* end of if not local EXT_INFO attribute record */
			} /* end of check for EXT_INFO attribute in attribute record */
		} ATTR_LIST_END;

	p_att = (idrp_attribute_record *) 0;
	return (p_att);
}
						
	
idrp_attribute_record *
create_ext_info_attr_rec(p_rth,p_rt)
rt_head		*p_rth;
rt_entry	*p_rt;
{
idrp_attribute_record 	*p_att;
int		size;		/* size of pdu */
int		type;		/* type of transfer bgp or plain external info */
u_char		*p_bgp;		/* universal pointer to bgp information for bgp 4 */ 


	/* should never come here with an null aspath pointer 
	 *  or a zero length aspath - check for it and warn if
	 * we found it
	 */

	if (!p_rt->rt_aspath)
		{
		trace (TR_IDRP,LOG_ERR,"idrp logic error - create_ext_info_attr_called with as path = 0");
		task_quit(p_rt->rt_aspath->path_id);
		}

	if (p_rt->rt_aspath->path_len == 0)
		{
		trace (TR_IDRP,LOG_ERR,"idrp logic error - create_ext_info_attr_called with as path len = 0");
		task_quit(p_rt->rt_aspath->path_id);
		}
		
	/*  fill in attribute record  with
	 *	1) rib_id - 
	 * 	2) idrp_mask -  with mandatory bits plus the ext-info 
	 *	3) idrp_local_proto - fill with this protocol
	 *	  (if other protocols share this structure - then set
	 *	   bits and add their idrpRoutes to the structure 
	 *	4) next_hop - from rt_entry
	 *	5) if route has AS_PATH,
	 *         then use to  1) set hop count
	 *			2) create cannonical RD path
	 *	6) fill in AS_PATH pointer
	 *	7) fill in ROUTE_ID with -1
         * 
	 */

	p_att = (idrp_attribute_record *) task_mem_malloc(idrp_this_node.task,sizeof(idrp_attribute_record));
	bzero(p_att,sizeof(idrp_attribute_record));
	p_att->rib_id = RIB_ID_DEFAULT;
	p_att->idrp_mask = IDRP_ATTBT_MANDATORY | IDRP_ATTBT_EXT_INFO;
	p_att->idrp_local_proto = p_rt->rt_gwp->gw_proto;
	p_att->next_hop = p_rt->rt_routers[0];

	/* fill in the RD_PATH and HOPCOUNT from AS_path structure */

	p_att->hopcount = 0;

#ifdef	IDRP_ASPATH

	/* we have an aspath structure 
	* - it would be nice to know who all has this structure
	* - is it just BGP-4 or is it IS-IS and all the other
	* - protocols.  It would be good if the gated could
	* - use this structure for any AS path configured or other
	* 	
	*  route has rt_aspath structure attached -
	*  so build EXT_INFO from that structure
	*  Be Aware O reader of IDRP that
	*  that the AS path is about to enter
	*  the transporter beam of a ISO enterprise.
	*  The AS pathway will be transformed in
	*  (not in an instant, but through lots of code)
	*  to RDI pathway.
	*
	*  Three types of transfer are currently planned
	* 	local_ext-info 	
	*		For ISO - just  IS-IS, static or ES-IS
	*		For IP - RIP, IS-IS which do not pass ASs
 	* --- these should not reach here
	*	IDRP_ASPATH_EXT_INFO 
 	*		For ISO - IS-IS with EXT_INFO		 
	*		for IP - EGP or IS-IS or OSPF with ASs
	* -- these should have an AS path of 1 or greater
	*	IDRP_ASPATH_BGP_XFR -
	*		We are transferring between like protocols.
	*		So transfer lots of information.
	*		BGP4 and IDRP speakers care inter-changable for IP.	
	*		(or at least that is the hope for now)
	*  -- these should have full bgp4 AS_SET and SEQUENCES
 	* 
	* routines called;	
	* - the idrp_aspath_to_canon_rdpath
	*  1) creates and links canon idrp rd path
	*  2) creates and links normal RDpath from
	*	AS path sequence
	*
	*/
		
	 
	p_att->hopcount = p_rt->rt_aspath->path_len; 	
	type = IDRP_ASPATH_EXT_INFO;
	p_bgp = (u_char *) 0;

	if (p_rt->rt_gwp->gw_proto == RTPROTO_BGP)
		{
		/* here we allow for additional passage
		 * of parameters from the bgp-4 implementations
		 * - right now it is blank - but it
		 * - is a prototype for further developement
		 */
 
		type = IDRP_ASPATH_BGP_XFR;
		p_bgp = (u_char *)0;
		}
		
	if (!idrp_aspath_to_canon_rdpath(p_rt->rt_aspath,p_att,&idrp_this_node,type,p_bgp))
		{
		/* the pathway contains an AS Loop - 
	 	 * IDRP checks to make sure it doesn't propogate
		 * anyone else's mistakes 
		 * or it's own confusion
		 * - So now we dump what we can to warn the user of AS path
		 * loopity - loop 
		 */

		trace (TR_IDRP,0,"idrp_aspath_to_canon_rdpath - find RDI loop pattern route %A",
			p_rth->rth_dest);
		trace (TR_IDRP,0,"dumping loop as RDI path for local node ");
		idrp_dump_rdi_path(p_att->rd_path,&idrp_this_node);	
		p_att = (idrp_attribute_record *) 0;
		return (p_att);
		}
#endif
	/* set up route_ID as there but with length as zero for
	 * external route
	 * skh  -- Should I set up a fake pdu??? (3/12/93)
	 * skh --- rexamine this as we go along (3/12/93)
	 * 
	 */

	p_att->as_path = p_rt->rt_aspath;
	p_att->attrib[IDRP_ATTR_ROUTE_SEPARATOR].present = TRUE;
	p_att->attrib[IDRP_ATTR_ROUTE_SEPARATOR].flags = IDRP_ATTRFLG_ROUTE_SEPARATOR;
	p_att->attrib[IDRP_ATTR_ROUTE_SEPARATOR].length = 0;
	p_att->attrib[IDRP_ATTR_ROUTE_SEPARATOR].data = (u_char *) 0;
	return (p_att);

}


#ifdef	IDRP_ASPATH
 
/* create an idrp RDI canonical path from AS path
 * - if type to create is local EXT_INFO - just create
 *	attribute path  
 * - if type is BGP transfer -  
 * 
 */

int 
idrp_aspath_to_canon_rdpath(p_as,p_att,peer,type,p_bgp)
as_path			*p_as;
idrp_attribute_record	*p_att;
idrpPeer		*peer;
int			type;
u_char			*p_bgp;
{
idrp_canon_rdpath	*p_can;		/* rdpath loop pointer */ 
idrp_canon_rdpath	*p_can_head;	/* head of canon rdpath list */
idrp_canon_rdpath	*p_can_tail;	/* tail of canon rdpath list */
register byte		*cp;		/* byte pointer for path attribute storage */ 
struct iso_net_addr	*p_rdi;		/* pointer to rdi */
int			size = 0;	/* size of rdi path in bytes */ 
int			hopcount = 0;	/* hopcount of rdi path */
int			i;		/* loop counter */
int			as;		/* integer value for as */  

	if (type == IDRP_ASPATH_BGP_XFR)
		{
		/* code needed to make this bgp compliant */
 
		trace (TR_IDRP, 0, "IDRP not ready for BGP interaction yet - NO path created ");
		return (FALSE);
		}

		 
	/* pull as by AS off the as path and create RD path 
	 * in both the cannonical format and the normal format. 
	 * However, it concerns me there is no ability to pass on the 
 	 * origin of the path in IDRP.  Perhaps a IP specific user 
	 * related path
	 * - note by the time the code gets here
	 * - it has been check 
	 */
 
	p_can_head = p_can_tail = (idrp_canon_rdpath *) 0;
	cp = PATH_PTR(p_as);

        for (i = p_as->path_len/2; i > 0; i--)
		{
		/* grab memory for the cannonical rd path and clear structure */
 
		p_can = (idrp_canon_rdpath *) task_mem_malloc(peer->task,sizeof(idrp_canon_rdpath));
		p_can->p_next = (idrp_canon_rdpath *) 0;

		/* get as as integer so that you can call mapping routine for
	 	 * mapping to rdi prefix for as 
	 	 */
		
            	as = (int)(*cp++) << 8;
            	as += (int)(*cp++);
		p_rdi = idrp_as_rdi_map(as);

		/* RDI in hand (watch for this daring ISO bird)
		 * create the RDI from the AS in this pathway 
	  	 * - copy RDI and then AS bytes at the end 
	 	 */ 
		  
		bcopy(p_rdi->isoa_genaddr,p_can->rdi.isoa_genaddr,p_rdi->isoa_len);
            	bcopy(cp,&p_can->rdi.isoa_genaddr[(p_rdi->isoa_len)],IDRP_AS_SIZE);
		p_can->rdi.isoa_len = p_rdi->isoa_len + IDRP_AS_SIZE;

		/* keep running total on the size of the RDI path 
		 * that must go in the attribute pathway 
		 */
 
		size = size + p_can->rdi.isoa_len + IDRP_RDPATH_RDI_LENGTH_BYTE_LEN;   
		hopcount++;

		/* update my rd_path linked list 
		 * and fly every onward to the next as 
		 * as a bird into the sunset of ISO-IP wars 
		 * or the bird of paradise into trouble 
		 */
 
		if (!p_can_head)
			{
			p_can_head = p_can_tail = p_can;
			}
		else
			{
			p_can_tail->p_next = p_can;
			p_can->p_next = (idrp_canon_rdpath *) 0;
			p_can_tail = p_can;
			}
		}

	p_att->hopcount = hopcount;

	if (type == IDRP_ASPATH_EXT_INFO)
		{
		return(generic_ext_info(p_can_head,p_att,size,peer));
		}
	if (type == IDRP_ASPATH_BGP_XFR)
		{
		return(idrp_aspath_bgp_xfr(p_can_head,p_att,peer,size,p_bgp));
		}
	
}

int
generic_ext_info(p_can_list,p_att,peer,size)
idrp_canon_rdpath	*p_can_list;	/* rd path list */
idrp_attribute_record	*p_att;		/* attribute record */
idrpPeer		*peer;		/* idrp peer ext_info associate with */
int			size;		/* size of RDI structure */
{
idrp_canon_rdpath	*p_rdi;
u_char			*cp;
	
	/* create from order list the canonical list and the
	 * actual RDI path sequence in memory
 	 *  allocate memory the size of RDI list 
	 */ 

	p_att->attr_len = size + IDRP_LOC_EXT_INFO_ATTR_BASE_LEN; 
	p_att->p_attr_pdu = (u_char *) task_mem_malloc(peer->task, p_att->attr_len);
	
	/* now it's time to create the fake pdu out of task memory
 	 * - get the size of the attribute paths to be created 
 	 * - put in route_id of -1 - so it is local
	 * - put in RD_PATH copied from the cannonical  
	 * - put in hopcount
	 * handle the Route_id - must be changed later to sequence 
         */		 

	p_att->attrib[IDRP_ATTR_ROUTE_SEPARATOR].present = TRUE;
	p_att->attrib[IDRP_ATTR_ROUTE_SEPARATOR].flags = IDRP_ATTRFLG_ROUTE_SEPARATOR;
	p_att->attrib[IDRP_ATTR_ROUTE_SEPARATOR].data = cp; 
	p_att->attrib[IDRP_ATTR_ROUTE_SEPARATOR].length = IDRP_ROUTE_SEPARATOR_LENGTH;
	bzero(cp,IDRP_ROUTE_SEPARATOR_LENGTH);
	cp += IDRP_ROUTE_SEPARATOR_LENGTH;

	/* copy in the RD_HOPCOUNT next */
	
	p_att->attrib[IDRP_ATTR_HOP_COUNT].present = TRUE;
	p_att->attrib[IDRP_ATTR_HOP_COUNT].flags = IDRP_ATTRFLG_HOP_COUNT;
	p_att->attrib[IDRP_ATTR_HOP_COUNT].data = cp; 
	p_att->attrib[IDRP_ATTR_HOP_COUNT].length = IDRP_RD_HOP_COUNT_LENGTH;
	*cp = p_att->hopcount;
	cp++;

	/* now for the RDI pathway */
	
	for (p_rdi = p_can_list; p_rdi; p_rdi->p_next)
		{
		*cp = p_rdi->rdi.isoa_len;
		cp++;
		bcopy(&p_rdi->rdi.isoa_genaddr,cp,p_rdi->rdi.isoa_len);
		if (!link_rd_canon(p_rdi,p_att)) 
			{
			/* found duplicate RDI using AS mapping *
			 * big time problem - return by leaving rd_path  
		 	 * and setting size to zero 
			 */

			task_mem_free(peer->task,p_att->p_attr_pdu);
			idrp_dump_rdi_path(p_can_list,peer);
			idrp_free_canon_rdpath(p_can_list,peer);
			return (FALSE);
			} 
		cp = cp + p_rdi->rdi.isoa_len;
		}		
	return (TRUE);
}

int
idrp_bgp_within_RD(p_idrp_rt)
idrpRoute	*p_idrp_rt;
{
	/* if AS path exists  and the hop count
	 * greater than one - we have more than
	 * this RD as part of our path 
	 */
	if (p_idrp_rt->p_attr->as_path &&
		(p_idrp_rt->p_attr->hopcount > 1))
		{
		return (FALSE);
		}
	return (TRUE);
}	

idrp_aspath_bgp_xfr(p_can_list,p_att,peer,size,p_bgp)
idrp_canon_rdpath	*p_can_list;
idrp_attribute_record	*p_att;
idrpPeer		*peer;
int			size;
u_char			*p_bgp;
{
idrp_canon_rdpath	*p_rdi;
u_char			*cp;

	/* no code yet */

	trace (TR_IDRP,0,"idrp_aspath_bgp_xfr called and no support yet -  Oh no momma bunny");
	return (FALSE);   

}	

/* AS to RDI mapping
 * 	configure table needed.  Right now we code in 
 * 	definition in the idrp for ip specification
 *
 */

struct iso_net_addr *idrp_as_rdi_map(as)
int	as;
{
	/* right now just use the global RDI defined in IDRP for IP */
	/* in the future configure a table */
	
	return (&rt_as_rdi);
}	

int
idrp_bgp_within_rd(p_idrp_rt)
idrpRoute	*p_idrp_rt;
{

	/* for now always claim bgp is within RD */
 
	return (TRUE);
}

#endif 

idrpPeer *	
idrp_ext_info_peer(p_rt,p_idrp_rt)
rt_entry	*p_rt;
idrpRoute	*p_idrp_rt;
{

	/* this function is left as a routine because although
	 * right now it seems that all ext_info should be in
	 * the local node's idrp peer, I not really sure
	 * what the implications of bgp to/from idrp should
	 * be.  Let's get IDRP for FAA working and then
	 * come back to this
	 */

	return (&idrp_this_node);
} 
