#include "include.h"
#include "idrp.h"

/* idrp_rt_phase2.c - 
 * Phase 2 code for idrp handling
 *
 * idrp_to_gated_pref - generated gated preference 
 *  
 * Phase 2 decision routes
 * 
 * tie_break(p_idrp_rt1,p_idrp_rt2)  - general tie break
 * tie_break_iso(p_idrp_rt1,p_idrp_rt2) - tie break for iso nlri
 * tie_break_ip(p_idrp_rt1,p_idrp_rt2) - tie break for ip
 * 
 * - utilities for iso tie breaking 
 * 
 * iso_next_hop_compare(p_idrp_rt1,p_idrp_rt2)
 * iso_multi_exit_compare(p_idrp_rt1,p/tie 
 * 
 * 
 * %% general comments
 *
 * We have the following  entry points to the IDRP code:
 * idrp_process_update -> parse_update -> idrp_route_mod
 *			 and when done -> idrp_process_pdu_routes
 * 
 * idrp_process_pdu_routes -> calls phase 1 processing routines:
 * -> phase1_internal->ph1_int_add_route 
 *		     ->ph1_with_route
 *		     ->ph1_with_rep_route
 * -> phase1_external->ph1_ext_add_route
 *		     ->ph1_with_route
 *		     ->ph1_with_rep_route
 *	-> send_best_ext -> del_routes_best_ext
 *   
 * Before the routes are handed to gated, preferences for
 *  gated are set so that it can proform phase 2 processing:
 *    check for other route
 *    tie_break routine
 *    gated_pref routine 
 *   
 *  
 * idrp_flash - called to handle the list of new active routes
 * 		from the gated routing table.
 *              The phase 3 routes are only sent to external peers.
 * 
 * idrp_flash -> phase3_processing -> phase3_status_change
 * ->idrp_send_phase3_route -> idrp_del_ph3_routes
 * 
 * 
*/ 



/*  
 * Phase 2 decision routes
 * 
 * - phase 2 installation is done by gated
 *   phase two references are done route by route
 *   upon input
 */

/* 
 * idrp_to_gated_pref(idrp_pref)
 */

int 
idrp_to_gated_pref(p_rt)
idrpRoute	*p_rt;
{

/* simply set IDRP PREFERENCE for gated
 *  plus  the idrp pref *2
 *  - multiply of preference allows for tie break
 *  to be set 
 */

/* set -initial preference to local infinity */

int	pref = RTPREF_KERNEL_REMNANT;

	if (p_rt->peer->type == IDRP_PEER_INTERNAL)
		{
		pref = RTPREF_IDRP_INT +  p_rt->pref_rcvd*2;
		}
	if (p_rt->peer->type == IDRP_PEER_EXTERNAL)
		{
		pref = RTPREF_IDRP_EXT  + p_rt->pref_cal*2;
		}

	if (p_rt->peer->type == IDRP_PEER_LOCAL)
		{
		pref = RTPREF_IDRP_EXT + p_rt->pref_cal*2;
		}
	return(pref);
} 



/*
 * tie_break(idrpRoute*, idrpRoute*)
 *
 * Description: Tie break between two addresses.  Value returned
 * is ISO_ADDR_LESS_THAN if the first addr should be chosen and 
 * it's ISO_ADDR_GREATER_THAN if the second address should be chosen.
 * I think that any other return value (e.g. ISO_ADDR_EQUAL) is an
 * error.
 *
 * Discussion:  We should probably have some generalized (not ISO-
 * specific) return values, such as TIEBREAK_LHS if the left-hand
 * argument wins the tie-break, TIEBREAK_RHS if the right-hand argument
 * wins, and TIEBREAK_ERROR if neither wins.
 */

int tie_break(p_idrp_rt1,p_idrp_rt2)
idrpRoute	*p_idrp_rt1;
idrpRoute	*p_idrp_rt2;
{
	if (p_idrp_rt1->family != p_idrp_rt2->family)
		{
		trace (TR_IDRP,0,"bogus - tie break 2 familes peer1 %s peer 2 %s Route1 %s Route 2 %s",  \
			p_idrp_rt1->peer->name,p_idrp_rt2->peer->name,  \
			iso_ptoa(&p_idrp_rt1->nlri),iso_ptoa(&p_idrp_rt2->nlri));
	
		/* two different families, all comparisons off */
		/* give iso priority for now */

		if (p_idrp_rt1->family ==  AF_ISO)
			{
			return (PFX_ADDR_GREATER_THAN);
			}
		else
			{
			return (PFX_ADDR_LESS_THAN);
			}
		}

	/* at this point the families are the same
	 * so dispatch to the appropriate families
	 * tie breaking routine for idrp routes 
	 */
 

	if (p_idrp_rt1->family == AF_ISO)
		{
		return (tie_break_iso(p_idrp_rt1,p_idrp_rt2));
		}  
	if (p_idrp_rt1->family == AF_INET)
		{
		return (tie_break_ip(p_idrp_rt1,p_idrp_rt2));
		}

	/* if another family - just leave it */

	return (PFX_ADDR_EQUAL);
}

				

/* 
 * tie_break_iso(idrpRoute*, idrpRoute*)
 *
 * Description: Tie break between two ISO addresses.  Value returned
 * is ISO_ADDR_LESS_THAN if the first addr should be chosen and 
 * it's ISO_ADDR_GREATER_THAN if the second address should be chosen.
 * I think that any other return value (e.g. ISO_ADDR_EQUAL) is an
 * error.
 */
int tie_break_iso(p_idrp_rt1,p_idrp_rt2)
idrpRoute	*p_idrp_rt1;
idrpRoute	*p_idrp_rt2;
{
int 	res = 0;

 	/* section 7.16.2.1 IDRP specification 
	 *  1) hop count compared  
	 *  2) MULTI_EXIT_DISC
	 *  3) compare NETs of peer sending route 
	 */

	if (p_idrp_rt1->p_attr == p_idrp_rt2->p_attr)
		{
		/* totally path attribute - so compare NETs */
		/* %%%% make sure this compare fits the description in 7.16.2.1 */
		return (compare_iso_addr(&p_idrp_rt1->peer->neighbor,&p_idrp_rt2->peer->neighbor));
		}

	if ((p_idrp_rt1->p_attr->idrp_mask & p_idrp_rt2->p_attr->idrp_mask) != p_idrp_rt1->p_attr->idrp_mask)
		{
		/* Not even the same path attribute list - do NET compare */ 
		/* %%%% make sure this compare fits the description in 7.16.2.1 */

		return (compare_iso_addr(&p_idrp_rt1->peer->neighbor,&p_idrp_rt2->peer->neighbor));
		}

	/* does the path diff only in Next hop and multi-exit discriminator */
 
	res = idrp_dif_path(p_idrp_rt1->p_attr,p_idrp_rt2->p_attr);

	if ((res & IDRP_ATTBT_NEXT_HOP) == res)
		{
		/*  only - next hop - compare next hops */

		return (iso_next_hop_compare(p_idrp_rt1->p_attr,p_idrp_rt2->p_attr));
		}

	if ((res & (IDRP_ATTBT_NEXT_HOP | IDRP_ATTBT_MULTI_EXIT_DISC)) == (IDRP_ATTBT_NEXT_HOP | IDRP_ATTBT_MULTI_EXIT_DISC))
		{
		/* only next hop and multi_exit */

		if (idrp_this_node.multi_exit) 
			{
			res = multi_exit_compare(p_idrp_rt1->p_attr,p_idrp_rt2->p_attr);
			if (res != ISO_ADDR_EQUAL)
				{
				return(res);
				}
			}
		}
	
	/* if multi_exit does not decide or is not on  - compare neighbor addresses */
	/* or if the paths differ by more than NEXT_HOP And MULTI_EXIT */
			
	res = compare_iso_addr(&p_idrp_rt1->peer->neighbor,&p_idrp_rt2->peer->neighbor);
	return (res);
}
	


int tie_break_ip(p_idrp_rt1,p_idrp_rt2)
idrpRoute	*p_idrp_rt1;
idrpRoute	*p_idrp_rt2;
{
int 	res = 0;
	/* ignore for now */
	return (res);
}	
int iso_next_hop_compare(p_att1,p_att2)
idrp_attribute_record *p_att1;
idrp_attribute_record *p_att2;
{
/* code this later */

return (1);
}
int multi_exit_compare(p_att1,p_att2)
idrp_attribute_record *p_att1;
idrp_attribute_record *p_att2;
{
/* code this later */
return (1);
}
