#include "include.h"
#include "idrp.h"

/* idrp_rt_phase3.c 
 *  - phase 3 routing table code  
 * 
 * Phase 3
 * 
 * idrp_flash(task,rtl) - receive all the active
 *		       routes from gated (LOC_RIB)
 *
 * phase_3(rtl) - phase 3 processing
 * idrp_phase3_dump(peer,rtl[]) - phase 3 processing initial peer up 
 *
 * phase3_status_change - change of status
 *
 * ph3_status_case1 - new idrp route
 * ph3_status_case2 - new active non-idrp route
 * ph3_status_case3 - old active idrp to new active idrp
 * ph3_status_case4 - old active ext_info to new active ext_info
 * ph3_status_case5 - old ext_info to new idrp
 * ph3_status_case6 - old active idrp to new active ext_info
 * ph3_status_case7 - old active idrp to delete route
 * ph3_status_case8 - old active ext_info to no new active
 *
 * idrp_send_phase3_routes
 * 
 *  
 * idrp_flash - called to handle the list of new active routes
 * 		from the gated routing table.
 *              The phase 3 routes are only sent to external peers.
 * 
 * idrp_flash -> phase3_processing -> phase3_status_change
 * ->idrp_send_phase3_route -> idrp_del_ph3_routes
 * 
 * idrp_newpolicy -> phase3_newpolicy ->add idrp routes
 * if any peers up, send out routes (idrp_send-phase3_route)
 * if all peer down, wait for phase3_dump to get it. 
* 
 *
 */ 

	
/* 
 * 
 * Phase 3
 *
 *
 * idrp_flash(task,rtl) - receive all the active
 *		       routes from gated (LOC_RIB)
 *
 * phase_3(rtl) - phase 3 processing
 *  
 * phase3_int(peer,rtl); - phase three processing for internal
 *			   peer (only on initial turn up)
 * phase3_ext(peer,rtl); - phase three processing for external peers
 * 
 *  dist_policy(send_list,peer) - run distribution policy
 *                                on the send_list for this peer
 *	returns modified send list
 *
 *  DIST(idrpRoute,peer) - run policy on a NLRI and peer 
 */ 	


/* idrp_flash - route called by gated when new idrp routes */ 

void
idrp_flash(tp,rtl)
task *tp;
rt_list *rtl;
{

     /* we get a flash update from the gated process 
      * walk the gated change list to find the update
      */ 

     trace(TR_IDRP | TR_TASK ,0,"idrp_do_flash: Doing flash update for IDRP");

    /* do processing, but do not send routes out
     * unless a peer is established */

	phase3_process(rtl);
}

void
phase3_process(rtl)
rt_list *rtl;
{
idrp_ann_list *p_idrp_ann;

	/* create the send lists from the rt_list */

	trace (TR_IDRP | TR_UPDATE, 0,"idrp flash processing called ");
	trace (TR_IDRP | TR_UPDATE, 0, "idrp status change routine called"); 

	p_idrp_ann = task_mem_malloc(idrp_this_node.task,sizeof(idrp_ann_list));
	bzero(p_idrp_ann,sizeof(idrp_ann_list));

	phase3_status_change(rtl,p_idrp_ann);

    	if (idrp_n_established == 0)
		{
		return;
		}

	if (idrp_ann_list_empty(p_idrp_ann))
		{
		trace (TR_IDRP | TR_UPDATE, 0, "idrp status change gave no announcements "); 
		return;
		}

	trace (TR_IDRP | TR_UPDATE, 0, "idrp status change routine finished with announcements"); 
	
	idrp_minadv_annlist(p_idrp_ann);	
	
	trace (TR_IDRP | TR_UPDATE,0,"idrp sending routes to peers "); 

	idrp_send_phase3_routes(p_idrp_ann);

	trace (TR_IDRP | TR_UPDATE, 0,"idrp sent routes to peers "); 
	trace (TR_IDRP | TR_UPDATE, 0,"idrp deleting routes requested DELETE after SENT");

	idrp_del_sent_routes(p_idrp_ann);

	if (idrp_set_min_route_timer())
		{
		trace (TR_IDRP | TR_UPDATE,0,"idrp set min_route advertisement timer");
		}

	/* free the announce list space */

	free_ann_list(p_idrp_ann);	
	task_mem_free(idrp_this_node.task,p_idrp_ann);	
}


int
idrp_phase3_dump(peer,actlist)
idrpPeer	*peer;
rt_list		*actlist[];
{
int		i;		/* loop counter NLRI_FAMILY_LOOP */
idrp_ann_list	ann_list;	/* announce list */
rt_entry	*p_rt;		/* gated route entry */	
idrpRoute	*p_irt;		/* idrpRoute pointer */	
u_int		rtbit;		/* bit to set in route when announcing other
				 * protocols route 
				 */

rt_head		*rth;		/* header of destination in gated */
rt_list		*rtl;		/* active route list */ 
idrp_ann_list	*p_ann_list;	/* pointer for announce list for loop */
idrp_send_list	*p_send_list = 0;	/* send list */
idrp_ann_list	*p_atl;		/* pointer for announce list for loop */


	/* this sets up the default RIB in send and announce list by
	 * default -  In QOS we will need to change this to be specific.
	 *
	 */

	bzero(&ann_list,sizeof(idrp_ann_list));
	rt_open(idrp_this_node.task);
	rtbit = idrp_this_node.task->task_rtbit;
	NLRI_FAMILY_LOOP(i)
		{
		rtl = actlist[i];
		RT_LIST(rth,rtl,rt_head) {

			if (socktype(rth->rth_dest) != nlri_family[i].family)
				{
				/* not AF_ISO or AF_INET should be here */
				continue;
				}

			p_rt = rth->rth_active;

			/* check for idrp route
			 * - all external routes
			 * will have been translated to
			 * idrp routes for sending out
			 * by the idrp_flash routines 
			 */

			if (p_rt->rt_gwp->gw_proto != RTPROTO_IDRP)
				{
				p_irt = idrp_find_ext_info_gated_rt(rth,p_rt);
				if (!p_irt)
					{
					continue;
					}

				}
			else
				{
				p_irt = (idrpRoute *) p_rt->rt_data;
				}
			if (DIST(p_irt,peer,IDRP_DIST_PEER_UP_DUMP))
				{
				link_ann_list(p_irt,&ann_list,NLRI_LINKED_P_ANN);
				}
			} RT_LIST_END(rth,rtl,rt_head);
		} NLRI_FAMILY_LOOP_END;

	AGGR(peer,&ann_list);
	p_ann_list = &ann_list;

	/* build the send list - by attribute,
	 * by family type (AF_ISO or AF_INET)
	 * and finally by NLRI
	 *
	 * - ANN_LIST - builds attribute
	 * - NLRI_FAMILY_LOOP builds  protocol type 
	 * - ANN_ANN_LIST builds the NLRI into pdus 
	 */

	ANN_LIST(p_atl,p_ann_list)
		{
		NLRI_FAMILY_LOOP(i)
			{
			ANN_ANN_LIST_WALK(p_irt,i,p_atl)
				{
				/* link route to send list and 	
			 	* when I have enough for an UPDATE  BISPDU
			 	* send it
			 	*/

				p_send_list = send_nlri_attr(p_irt,p_atl->p_attr,p_send_list,peer);
				} ANN_ANN_LIST_WALK_END;
			} NLRI_FAMILY_LOOP_END;

		/* before you go to the next attribute record - flush this one's nlri's out */ 
		flush_att_send_list(peer,p_atl->p_attr,p_send_list);
		} ANN_LIST_END;	

	/* release the single send list once
	 * the routes have been sent - this will change
	 * into a linked list of send lists farily
	 * easily.
	 */

	free_send_list(p_send_list,peer);
	free_ann_list(p_ann_list);
	rt_close(idrp_this_node.task,&idrp_this_node.gw,0,NULL);
	trace (TR_IDRP,0, "phase 3 - first time dump for neighbor %s ",peer->name);
}

idrp_ann_list *
phase3_newpolicy(rtl,p_ann_list)
rt_list *rtl;
idrp_ann_list *p_ann_list;
{
idrpPeer *peer;
u_int	rtbit;
rt_head *rth;
rt_entry *p_old_rt;
rt_entry *p_rt;
rt_changes	*rtc;
int	changes = 0;

	bzero(p_ann_list,sizeof(idrp_ann_list));
	rt_open(idrp_this_node.task); 
	RT_LIST(rth,rtl,rt_head) {

		/* check to see if the family supported in idrp */

		if (!sock_supported(&idrp_this_node,rth->rth_dest))
			{
			continue;
			}

		/* unknown status that route gets into
		 * check for idrp bit set on active  
		 */ 

		p_rt = rth->rth_active; 
		if (!p_rt)
			{
			/* problem with the flash to new policy.  ignore this route
			 */
 
			trace (TR_IDRP,LOG_ERR,"idrp_newpolicy called with no active route nlri %A ",
				rth->rth_dest);
			continue;
			} 
		if (p_rt->rt_gwp->gw_proto == RTPROTO_IDRP)
			{
			/* new route protocol idrp */
			ph3_status_case1(rth,p_ann_list);
			}
		else
			{
			/* new route protocol other than idrp
			 */ 
	
			if (ph3_status_case2(rth,p_ann_list))
				{
				changes++;
				}
			}	

	} RT_LIST_END(rth,rtl,rt_head);
	rt_close(idrp_this_node.task,&idrp_this_node.gw,changes,NULL); 

	/* announce list is free in the newpolicy code after transmission
	 * of rib refreshes to all connected peers
	 */

	return(p_ann_list);

}

/* %% - phase 3 status change routines */

void
phase3_status_change(rtl,p_ann_list)
rt_list *rtl;
idrp_ann_list	*p_ann_list;
{
idrpPeer *peer;
u_int	rtbit;
rt_head *rth;
rt_entry *p_old_rt;
rt_entry *p_rt;
rt_changes	*rtc;
int	changes = 0;
int	not_changed;

	/* get an announce list allocated */

	rt_open(idrp_this_node.task);
	rtbit = idrp_this_node.task->task_rtbit;

	RT_LIST(rth,rtl,rt_head) {

		/* check for support for nlri passage for this
		 * family of routes
		 */
 
		if (!sock_supported(&idrp_this_node,rth->rth_dest))
			{
			continue;
			}


		/* route is ISO or INET */

		p_old_rt = rth->rth_last_active;
		p_rt = rth->rth_active; 

		if ((!p_rt) && (!p_old_rt))
			{
			continue;
			}
			  

		if (!p_old_rt)
			{
			/* no old route */ 
			if (p_rt->rt_gwp->gw_proto == RTPROTO_IDRP)
				{
				/* new route protocol idrp */
				ph3_status_case1(rth,p_ann_list);
				}
			else
				{
				/* new route protocol other than idrp */
				if ( ph3_status_case2(rth,p_ann_list))
					{
					changes++;
					}
				}
			}
		else
			{
			/* old route exists 
			 * - Is it simply a change to the same route?
			 * or is it a change or routes ?
			 */
			 	
			if (p_rt == p_old_rt)
				{
				/* change to existing same route */
				not_changed = 0;

				rtc = rth->rth_changes;
				if (rtc)
					{
					if (p_rt->rt_gwp->gw_proto == RTPROTO_IDRP)
						{
						/* old route idrp, new route idrp */
						ph3_status_case3(rth,p_ann_list);
						}
					else
						{
						/* old route ext_info, new route ext_info */

						if (ph3_status_case4(rth,p_ann_list))
							{
							changes++;
							}
						}	 
					}

				}
			else
				{	
				/* routes are different and change is real */

				if ((p_rt->rt_state & RTS_DELETE) == 0)
					{
					/* replacement of rt_entry route structures */
	
					if (p_old_rt->rt_gwp->gw_proto == RTPROTO_IDRP)
						{  
						if (p_rt->rt_gwp->gw_proto == RTPROTO_IDRP)
							{
							/* new route idrp and old route idrp */

							ph3_status_case3(rth,p_ann_list);
							}
						else
							{
							/* old route idrp, new route ext_info */

							if (ph3_status_case6(rth,p_ann_list))
								{
								changes++;
								}
		
							}
						}
					else
						{
						if (p_rt->rt_gwp->gw_proto == RTPROTO_IDRP)
							{
							/* old route ext_info, new route idrp */
							ph3_status_case5(rth,p_ann_list);
							}
						else
							{
							/* old route ext_info and ext_info  */
							if (ph3_status_case4(rth,p_ann_list))
								{
								changes++;
								}
							}
						}
					}
				
				else
					{
					/* current route is being deleted */
					if (p_old_rt->rt_gwp->gw_proto == RTPROTO_IDRP)
						{
						ph3_status_case7(rth,p_ann_list);
						}
					else
						{
						ph3_status_case8(rth,p_ann_list);
						}
					}

				}
			}

	} RT_LIST_END(rth,rtl,rt_head);
	rt_close(idrp_this_node.task,&idrp_this_node.gw,changes,NULL);
}


/* %% phase3 - status change routines
 * ph3_status_case1 - new idrp route
 * ph3_status_case2 - new active non-idrp route
 * ph3_status_case3 - old active idrp to new active idrp
 * ph3_status_case4 - old active ext_info to new active ext_info
 * ph3_status_case5 - old ext_info to new idrp 
 * ph3_status_case6 - old active idrp to new active ext_info
 * ph3_status_case7 - old active idrp to delete route
 * ph3_status_case8 - old active ext_info to no new active
 */
 

/* ph3 - status case 1 - new idrp route */
 
void 
ph3_status_case1(rth,p_ann_list)
rt_head		*rth;
idrp_ann_list	*p_ann_list; 
{
rt_entry 	*p_rt;
idrpRoute	*p_idrp_rt;

	/* in the  idrpRoute structure
	 * set LOC_RIB status
	 * link through announce list to annouce outward 
	 * add to min route advertisement timer list
	 */
  
	p_idrp_rt = (idrpRoute *) rth->rth_active->rt_data;
	trace (TR_IDRP|TR_UPDATE, 0, "case1: IDRP peer %s route %s flashed",p_idrp_rt->peer->name,iso_ptoa(&p_idrp_rt->nlri));

	IDRP_STATUS_BIT_SET(p_idrp_rt,IDRP_STATUS_LOC_RIB); 

	/* link through p_next to announce list */

	if (p_idrp_rt->peer->type == IDRP_PEER_EXTERNAL ||
		(p_idrp_rt->peer->type == IDRP_PEER_LOCAL))
		{
		/* only re-distribute LOC_RIB changes
	         * from idrp external peers or IDRP_PEER_LOCAL   
		 * ( LOC_RIB changes = gated active route changes)   
		 *
		 */

		if (!rtbit_isset(p_idrp_rt->p_rt,idrp_this_node.task->task_rtbit))  
			{
			rtbit_set(p_idrp_rt->p_rt,idrp_this_node.task->task_rtbit);
			}
		link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_ANN);	
		}
}


int
ph3_status_case2(rth,p_ann_list)
rt_head *rth;
idrp_ann_list	*p_ann_list; 
{
rt_entry	*p_rt;
idrpRoute	*p_idrp_rt;
idrpPeer	*peer;
u_int		rtbit;	
int		route_id_in;
int		pref;

	/* calculate preference to find out if
	 * it should be passed via IDRP route
	 */

	trace(TR_IDRP,0,"ph3_status_case2 entered"); 
	rtbit = idrp_this_node.task->task_rtbit;
	p_rt = rth->rth_active;
	pref = idrp_ext_info_pref(p_rt);
	if (!pref)
		{
		/* return no changes */
		return(FALSE);
		}

	/* try to find  an IDRP route
	 * generated from this route
	 * in gated tables
	 */

	p_idrp_rt = idrp_find_ext_info_gated_rt(p_rt->rt_head,p_rt);	
	if (p_idrp_rt)
		{
		if (verify_extinfo_idrp_rt(p_idrp_rt,p_rt))
			{
			link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_ANN);
			return(TRUE);
			}
		else
			{
			return(FALSE);
			}
		}

	p_idrp_rt = ph3_create_ext_route(p_rt,rth,pref);
	if (!p_idrp_rt)
		{
		return(FALSE);	
		}
	link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_ANN);

	trace (TR_IDRP,0,"ph3_status_case2 called peer %s route %s ", 
		p_idrp_rt->peer->name,iso_ptoa(&p_idrp_rt->nlri));
	return(TRUE);

}

void
ph3_status_case3(rth,p_ann_list)
rt_head *rth;
idrp_ann_list	*p_ann_list; 
{
rt_entry	*p_rt,*p_rt_old;
idrpRoute	*p_idrp_rt,*p_idrp_rt_old;


	/* logic for change of active route
	 * sub-case 1 - min route advertisement timer running  
	 *		or mininum route advertisement timer
	 *		running for RD
	 * Just set change flag and exit 
	 */

	p_rt = rth->rth_active;
	p_rt_old = rth->rth_last_active;
	p_idrp_rt = (idrpRoute *) p_rt->rt_data;
	p_idrp_rt_old = (idrpRoute *) p_rt_old->rt_data;

	/* check the status on the new route
	 * - I do not think it should be running the MIN_ADV timer
	 * - until we send it out once 
	 * -- so complain here */
 

	if ((p_idrp_rt->status & IDRP_STATUS_MIN_ADV_RUN)
		|| (p_idrp_rt->status & IDRP_STATUS_MIN_RDADV_RUN))
		{
		IDRP_STATUS_BIT_SET(p_idrp_rt,IDRP_STATUS_MIN_ADV_CHG);
		trace(TR_IDRP,LOG_ERR," route %s peer %s has min advertisement set on new route ", 
			iso_ptoa(&p_idrp_rt->nlri),p_idrp_rt->peer->name);	
		}

	if ((p_idrp_rt_old->status & IDRP_STATUS_MIN_ADV_RUN)
		|| (p_idrp_rt_old->status & IDRP_STATUS_MIN_RDADV_RUN))
		{
		IDRP_STATUS_BIT_SET(p_idrp_rt,IDRP_STATUS_MIN_ADV_CHG);
		if ((p_idrp_rt_old->status & IDRP_STATUS_WITH) || 
			(p_idrp_rt_old->status & IDRP_STATUS_DELETE))
			{
			link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_WITH);
			}
		return;
		}

	/* subcase 2 - No minimum advertisement timer
	 *		old route deleted 
	 */

	if (p_idrp_rt->status & IDRP_STATUS_DELETE)
		{ 
		/* link the new route to the idrp list */
		link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_ANN);
	
		/* delete the gated route - although
		 * we should do two deletes
		 * but this one should clear
		 * the gated bits and having
		 * cleared all other bits
		 * release the rt_entry
		 */
			
		idrp_del_rt_gated(p_idrp_rt_old);
		idrp_free_outlist(p_idrp_rt_old);
		/* think this is P_NEXT_NLRI - but watch out here
		 * to make sure no P_WITH links
		 */

		idrp_free_nlri_att_rec(p_idrp_rt_old,NLRI_LINKED_P_NEXT);
		idrp_free_idrpRoute(p_idrp_rt_old);
		return;
		}


	/* subcase - 3 
	 * No delete, no min route advertisement timer
	 * - just a straight replace of one idrp route
	 * with another
	 * - just link the idrp route to the announce list
	 */

	link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_ANN);
	return;

}

/* old ext_info to new ext_info route
 * test: old_route = protocol other than idrp
 *       new_route = protocol other than idrp 
 */


int
ph3_status_case4(rth,p_ann_list)
rt_head *rth;
idrp_ann_list	*p_ann_list; 
{

rt_entry	*p_rt,*p_rt_old;
idrpRoute	*p_idrp_rt = 0;
idrpRoute	*p_idrp_rt_old = 0;
u_int		rtbit;
int		pref = 0;

	/* need to print out definition here -- jgs help needed */

	trace (TR_IDRP,0,"ph3_status_case 4 called destination ");

	/* set-up the old route and the new route */

	p_rt = rth->rth_active;
	p_rt_old = rth->rth_last_active;
	rtbit = idrp_this_node.task->task_rtbit;


	/* Do pref on new route */

	pref = idrp_ext_info_pref(p_rt);

	/* check to see if old route has idrpRoute */

	p_idrp_rt_old = idrp_find_ext_info_gated_rt(rth,p_rt_old);
	if (p_idrp_rt_old)
		{
		if (!verify_extinfo_idrp_rt(p_idrp_rt_old,p_rt_old))
			{
			/* bit set on the gated route for 
			 * but a problem - set p_idrp_rt_old to null
			 */
			p_idrp_rt_old = (idrpRoute *) 0;
			}
		}

	if (pref)
		{
		/* see if you have any IDRP route for the
		 * new active route
		 */ 

		p_idrp_rt = idrp_find_ext_info_gated_rt(rth,p_rt);
		if (!p_idrp_rt)
			{
			if (!verify_extinfo_idrp_rt(p_idrp_rt,p_rt))
				{
				trace(TR_IDRP,LOG_ERR," ph3_status_case4: found idrp rt but failed verify nlri %s", 
					iso_ptoa(&p_idrp_rt->nlri)); 
				task_quit(0);
				}
			}
		else
			{
			p_idrp_rt = ph3_create_ext_route(p_rt,rth,pref);
			}
		}

	/* now for the sub cases tests */

	/* subcase 1 - old ext_info route did not have idrp route
	 *	       new ext_info route does not have idrp route
	 */

	if (!p_idrp_rt_old)
		{
		if (!pref)
			{
			return(FALSE);
			}
		else
			{
			/* subcase 2 - old ext_info route did not have idrp route
			 * new ext_info route has idrp route - simply link to 
		 	 * announce list
			 */
			link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_ANN);
			return(TRUE);
			}
		} 

	/* all logic beneath hear has hold routes of IDRP */

	if (!pref)
		{
		if (!p_idrp_rt_old->p_min_advRD)
			{
			/* subcase 3 - old ext_info route had idrp route
			 *		new ext_info route has no idrp route
			 *		Minimum Route Advertisement timer not set in old idrp route
			 */

			IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_DEL_SEND);
			link_ann_list(p_idrp_rt_old,p_ann_list,NLRI_LINKED_P_WITH);
			return(FALSE);
			}
		else
			{	
		
			/* subcase 4 - old ext_info route had idrp route
			 *		new ext_info route has no idrp route set-up
			 * 	        Min Route Advertisement this RD set in old idrp route
			 */

			IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_WITH);
			IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_MIN_ADV_CHG);
			link_ann_list(p_idrp_rt_old,p_ann_list,NLRI_LINKED_P_WITH);
			return(FALSE);
			}
		}

	/* all routes under here have preference set - so new route was created 
	 * subcase 5 and 6
	 */

	if (p_idrp_rt_old->p_min_advRD)
		{
		/* old route has min route advertisement set
		 */
 
		IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_WITH);
		IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_MIN_ADV_CHG);
		}
	else
		{
		link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_ANN);
		}
	return(TRUE);
}

/* old_ext_info -> new idrp route
 * test: old_route = non_idrp
 *       new_route = idrp
 */


void
ph3_status_case5(rth,p_ann_list)
rt_head *rth;
idrp_ann_list	*p_ann_list; 
{
rt_entry	*p_rt,*p_rt_old;
idrpRoute	*p_idrp_rt = 0;
idrpRoute	*p_idrp_rt_old = 0;
u_int		rtbit;

	trace (TR_IDRP,0,"ph3_status_case5 called");

	/* set-up the old route and the new route */

	p_rt = rth->rth_active;
	p_rt_old = rth->rth_last_active;
	rtbit = idrp_this_node.task->task_rtbit;
	p_idrp_rt = (idrpRoute *) p_rt->rt_data;

	if (!rtbit_isset(p_rt_old,rtbit))
		{
		link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_ANN);
		return;
		}

	/* possible old idrp route */

	
	p_idrp_rt_old = idrp_find_ext_info_gated_rt(rth,p_rt_old);
	if (p_idrp_rt_old)
		{
		if (!verify_extinfo_idrp_rt(p_idrp_rt_old,p_rt_old))
			{
			/* bit set on the gated route for 
			 * but a problem - set p_idrp_rt_old to null
			 * - so stop the prototype for now
			 */
			task_quit(0);
			}
		}
	if (p_idrp_rt_old = p_idrp_rt)
		{
		trace(TR_IDRP,0,"IDRP ROUTE from EXT_INFO should not be active route nlri %s", 
				iso_ptoa(&p_idrp_rt->nlri));	
		task_quit(0);
		}
 
	if (p_idrp_rt->status & IDRP_STATUS_MIN_ADV_RUN)
		{
		IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_MIN_ADV_CHG);
		}
	else
		{
		link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_ANN);
		}
	
	return;
	
}


/* status case 6  - active idrp route -> new ext_Info route
 * test: old route = idrp
 *       new route = ext_info
 */

int
ph3_status_case6(rth,p_ann_list)
rt_head *rth;
idrp_ann_list	*p_ann_list; 
{
rt_entry	*p_rt,*p_rt_old;
idrpRoute	*p_idrp_rt = 0;
idrpRoute	*p_idrp_rt_old = 0;
u_int		rtbit;
int		pref;

	trace (TR_IDRP,0,"ph3_status_case6 called");

	/* set-up the old route and the new route */

	p_rt = rth->rth_active;
	p_rt_old = rth->rth_last_active;
	rtbit = idrp_this_node.task->task_rtbit;
	p_idrp_rt_old = (idrpRoute *) p_rt_old->rt_data;


	/* Do pref on new route */

	pref = idrp_ext_info_pref(p_rt);

	/* check to see if old route has idrpRoute */

	if (pref)
		{
		/* see if you have any IDRP route for the
		 * new active route
		 */ 

		p_idrp_rt = idrp_find_ext_info_gated_rt(rth,p_rt);
		if (!p_idrp_rt)
			{
			if (!verify_extinfo_idrp_rt(p_idrp_rt,p_rt))
				{
				trace(TR_IDRP,LOG_ERR," ph3_status_case4: found idrp rt but failed verify nlri %s", 
					iso_ptoa(&p_idrp_rt->nlri)); 
				task_quit(0);
				}
			}
		else
			{
			p_idrp_rt = ph3_create_ext_route(p_rt,rth,pref);
			}
		}

	else
		{
		/* subcases */

		/* subcase 1 - old idrp route going away
		 * 	     - no new route (pref = 0)
		 * 	     - min route advertisement timer running
	 	 * set the min_adv_chg and link withdrawl to idrp list
		 */

		if (p_idrp_rt->status & IDRP_STATUS_MIN_ADV_RUN)
			{
			IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_MIN_ADV_CHG);
			}

		/* subcase 2 - old idrp route going away
	 	 * no new route (pref = 0
		 * min route adv timer not running 
		 * - still link withdrawl to min adv list
		 */


		link_ann_list(p_idrp_rt_old,p_ann_list,NLRI_LINKED_P_WITH);
		return;
		}
		

	/* beyond her prefer is set to create new route */

	/* subcase 3 - old idrp route
	 *	     - new route,
	 *	      -  min adv timer runnning
	 */
	if (p_idrp_rt_old->status & IDRP_STATUS_MIN_ADV_RUN)
		{
		IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_MIN_ADV_CHG);
		return;
		}
	else
		{
		/* subcase 4 - old idrp route 
		 * 	     - new idrp route ext_info
		 *	     - no min adv timer running
		 */
		link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_ANN);
		}

	 
	
}

/* delete of active idrp route
 * - no new route 
 */

void
ph3_status_case7(rth,p_ann_list)
rt_head *rth;
idrp_ann_list	*p_ann_list; 
{
rt_entry	*p_rt_old;
idrpRoute	*p_idrp_rt_old = 0;
u_int		rtbit;

	trace (TR_IDRP,0,"ph3_status_case7 called");

	/* set-up the old route and the new route */

	p_rt_old = rth->rth_last_active;
	rtbit = idrp_this_node.task->task_rtbit;
	p_idrp_rt_old =(idrpRoute *)  p_rt_old->rt_data;

	/* delete of IDRP route and Minimum Advertisement timer running */

	if (p_idrp_rt_old->status & IDRP_STATUS_MIN_ADV_RUN)
		{
		IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_MIN_ADV_CHG);
		IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_DELETE);
		return;
		}

	/* no Minimum advertisement timer running */

	IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_DEL_SEND);
	link_ann_list(p_idrp_rt_old,p_ann_list,NLRI_LINKED_P_WITH);
	return; 	
}

/* Status case 8 :
 *  old active route = ext_info (deleted),  no new active route
 * 	
 */

void
ph3_status_case8(rth,p_ann_list)
rt_head *rth;
idrp_ann_list	*p_ann_list; 
{
rt_entry	*p_rt_old;
idrpRoute	*p_idrp_rt_old = 0;
u_int		rtbit;


	trace (TR_IDRP,0,"ph3_status_case8 called");

	/* set-up the old route and the new route */

	p_rt_old = rth->rth_last_active;
	rtbit = idrp_this_node.task->task_rtbit;


	/* does this ext_info route have idrp route bit
	 * set? - if yes, continue.
	 * if no, exit routine - no processing if no idrp route
	 */
 
	if (!rtbit_isset(p_rt_old,rtbit))  
		{
		return;
		}

	/* withdraw logic for idrp route created out
	 * of ext_info
	 */ 

	p_idrp_rt_old = idrp_find_ext_info_gated_rt(rth,p_rt_old);
	if (p_idrp_rt_old)
		{
		if (!verify_extinfo_idrp_rt(p_idrp_rt_old,p_rt_old))
			{
			/* bit set on the gated route for 
			 * but a problem - set p_idrp_rt_old to null
			 * - so stop the prototype for now
			 */
			task_quit(0);
			}
		}
	IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_WITH);
	link_ann_list(p_idrp_rt_old,p_ann_list,NLRI_LINKED_P_WITH);

	/* if minimum route advertisement timer running
	 * - set change and delete.
	 * timer routine will clean up this route
	 */ 
 
	if (p_idrp_rt_old->status & IDRP_STATUS_MIN_ADV_RUN)
		{
		IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_MIN_ADV_CHG);
		IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_DELETE);
		}
	else
		{
		IDRP_STATUS_BIT_SET(p_idrp_rt_old,IDRP_STATUS_DEL_SEND);
		}

}

void 
idrp_send_phase3_routes(p_ann_list)
idrp_ann_list *p_ann_list;
{
idrpPeer		*peer;		/* peer structure pointers */
idrpRoute		*p_irt;		/* pointer to idrpRoute for loop processing */ 
idrp_ann_list		*p_atl;		/* announce list */
idrp_attribute_record	*p_att_dist;	/* distribution attribute record pointer */
idrp_send_list		*p_send_list = 0;	
int			i;		/* NLRI loop counter */ 


	/* trace the face are trying to send phase 3 routes */

	trace( TR_IDRP,0,"idrp_send_phase3_routes called");

	IDRP_EXT_PEER(peer)
		{
		/* walk through the announce list building
	 	 * send lists and sending them to the external peers 
		 */
		if ((peer->idrp_flags & IDRPF_CONNECT) == 0)
			{
			/* if peer not established - don't send the route */
			continue;
			}
 
		ANN_LIST(p_atl,p_ann_list)
			{
			p_att_dist = DIST_ATTR(p_atl->p_attr,peer);
			NLRI_FAMILY_LOOP(i)
				{
				ANN_WITH_LIST_WALK(p_irt,i,p_ann_list)
					{
					p_send_list = send_with_attr(p_irt,p_att_dist,p_send_list,peer);
					} ANN_WITH_LIST_WALK_END;

				ANN_ANN_LIST_WALK(p_irt,i,p_ann_list)
					{
					if (p_att_dist)
						{
						if (DIST(p_irt,peer,IDRP_DIST_FLASH))
							{
							p_send_list = send_nlri_attr(p_irt,p_att_dist,p_send_list,peer);
							}
						}
					} ANN_ANN_LIST_WALK_END;
				} NLRI_FAMILY_LOOP_END;
			flush_att_send_list(peer,p_att_dist,p_send_list);
			} ANN_LIST_END;
		free_send_list(p_send_list,peer);
		} IDRP_EXT_PEER_END;

}	

int
verify_extinfo_idrp_rt(p_idrp_rt,p_rt)
idrpRoute	*p_idrp_rt;
rt_entry	*p_rt;
{
	if (rtbit_isset(p_rt,idrp_this_node.task->task_rtbit))  
		{
		/* idrp route exists for this route 
		 * so go find it and link it to 
		 * announce list
		 */

		if (p_idrp_rt)
			{
			trace (TR_IDRP,0,"destination route %s has idrp announce bit set, with idrp route match", 
				iso_ptoa(&p_idrp_rt->nlri));
			return(TRUE); 
			}
		else
			{
			/* --- skh (3/10/93) Need to talk to John -where is the sockun_toa replacement code
			 *
			trace (TR_IDRP,LOG_ERR,"destination route %s has idrp announce bit set, with no idrp route", 
				sockun_toa(p_rt->rt_dest));
			*/
			}
		return(FALSE);
		} 
	else
		{
		if (p_idrp_rt)
			{
			/* route already has idrp route - logic problem */
			trace (TR_IDRP,0,"ph3_status_case2 -logic error route %A already has idrp route %s", 
				p_rt->rt_dest,iso_ptoa(&p_idrp_rt->nlri));	
			return(FALSE);
			}
		}
}

/* create an external route for phase 3 */


idrpRoute *
ph3_create_ext_route(p_rt,rth,pref)
rt_entry	*p_rt;
rt_head		*rth;
int		pref;
{
idrpRoute_entry	*p_rte,*p_rte_last;
int		family;
idrp_attribute_record *p_att;
idrpRoute		*p_idrp_rt;
idrpPeer		*peer;
int			route_id_in;


	/* set up the family for this destination */

	family = socktype(rth->rth_dest);

	/* No idrp route exists for this route
	 * - so set-up an IDRP route for the
 	 *  this gated route.
	 * - check for route in the  
	 *  1.) set the IDRP protocol in the rt_entry
	 */

	rtbit_set(p_rt,idrp_this_node.task->task_rtbit);
 
	/* If pref is non-zero we will accept the
	 * route and there is no route already.
	 * create a route type based on the
	 * the value of the preference flag
	 * 
	 * pref is  
	 * IDRP_EXT_INFO_IGP -
	 * create a non-idrp route with
	 * the local attribute record
	 * for non-exterior routes
	 *   
	 * if pref is
	 * IDRP_EXT_INFO_EXT_INFO 
	 * have an IDRP route with the EXT_INFO
	 * attribute. 
	 */

	if (pref == IDRP_EXT_INFO_IGP)
		{
		/* gated route from this IGP (IS-IS)
	 	 * for now all external routes are 
		 * set-up with no options on external
		 * attribute
		 * this will change in the
		 * future as we
		 * work on it per configuration
		 */ 

		p_att = find_local_att_rec(p_rt->rt_routers[0],(idrpRoute_options *) 0);
		}
	else
		{
		/* pref = IDRP_EXT_INFO_EXT_INFO */
	
		p_att = find_ext_info_attr_rec(rth,p_rt);
		if (!p_att)
			{ 
			/* create an external info attribute record
			 * - if bgp or as path this will translate
			 * - in	to rdi paths
			 */ 

			p_att = create_ext_info_attr_rec(rth,p_rt);
			}
		}

	/*
	 * 3) set the peer to the peer from this attribute record 
	 *	if non-bgp:
         *		this will be local_nodes's idrp peer
	 *      if BGP: may change may transfer the peer this
	 *		to an IDRP parallel peer structure	 
	 * 
	 */

	peer = idrp_ext_info_peer(p_rt,p_idrp_rt);

	/*
	 *  4.) create an idrpRoute for this Route
	 *	 - allocate the idrp route and stuff in initial info
	 *	 - put nlri in route
	 *	 - attach this route to the attribute record 
	 *	   (will be local node except for bgp 
	 *
	 * 5) set the Route_id_in to be the external info route_id
	 *		information  is received one 
	 *
	 * 	we build the attribute record with the right peer
	 * 	this double check is useful 
	 */

	if (pref == IDRP_EXT_INFO_IGP)
		{
		route_id_in = IDRP_LOCAL_ROUTE_ID;
		}
	else	
		{
		route_id_in = IDRP_EXT_INFO_ROUTE_ID;
		}

	p_idrp_rt = idrp_alloc_idrpRoute(peer,p_att,family,route_id_in);
	sockun_to_prefix(p_rt->rt_dest,p_rt->rt_dest_mask,&p_idrp_rt->nlri);	

	/* here we need to fill in the idrp_rt allocated
	 * 6) fill in rt_entry as zero until one gets assigned by gated
	 * 7) fill EXT_INFO portion of route to point to gated route
	 * 8) p_loc_rt_peer  - does it get filled in with the
	 *		      peer structure of the exterior router
	 *		      that sent the route or the next hop
	 *		      - next hop is set right now from 
	 *		      - the attribute record 
	 */


	p_idrp_rt->p_rt = (rt_entry *) 0; 
	p_idrp_rt->p_rt_ext_info = p_rt;
	p_idrp_rt->p_loc_rt_peer = p_rt->rt_routers[0];
	p_idrp_rt->status = IDRP_STATUS_LOC_RIB & IDRP_STATUS_EXT_INFO;
	

	/* 9) link the idrp route to the attribute record
	 *    with EXT_INFO_ROUTE_ID and this local node 
	 */ 

	p_rte = idrp_find_routeid_chain(p_idrp_rt,&p_rte_last);
		
	if (p_rte)
		{
		/* found it - add the new idrpRoute to
		 * idrpRoute_entry in chain with EXT_INFO
		 */

		if (!add_nlri_routeid_chain(p_rte,p_idrp_rt))
			{
			trace (TR_IDRP,0,"error in finding routeid chain route %s peer %s", 
				iso_ptoa(&p_idrp_rt->nlri),p_idrp_rt->peer); 
			idrp_free_idrpRoute(p_idrp_rt);
			return(0);
			}
		}
	else
		{
		/* none found -  create one based on the route ID
	 	 * and add it to the route_id chain    
		 */

		p_rte = (idrpRoute_entry *) task_mem_malloc(idrp_this_node.task,sizeof(idrpRoute_entry));
		bzero(p_rte,sizeof(idrpRoute_entry));
		p_rte->peer = peer;
		p_rte->route_id = p_idrp_rt->route_id_in;
		p_rte->status = IDRP_ATTR_ROUTE_ID_NORM;
		p_rte->p_route[family_to_nlri_id(p_idrp_rt->family)] = p_idrp_rt;
		p_idrp_rt->p_next_nlri = (idrpRoute *) 0;
		}	

	/* 10) increment reference count on the attribute record
	 * 11) here we need to calculate the idrp_preference
	 */

	p_att->ref_cnt++;	
	idrp_pref(p_idrp_rt);

	/*
	 * 12) add route to gated 
	 *     - call routine to check gated for the idrp
	 * 	 route prior to adding the route
	 *	to gated.  if valid continue, if not
	 *	release this route.
	 */

	if(!idrp_add_local_rt_gated(p_idrp_rt))
		{
		/* already have this idrp route -
		 * so release this route structure
		 * and nlri from attribute list
		 */
		trace (TR_IDRP,0,"ph3_status_case2 called, ignored route %s", 
			iso_ptoa(&p_idrp_rt->nlri));
		idrp_free_nlri_att_rec(p_idrp_rt,NLRI_LINKED_P_NEXT); 
		idrp_free_idrpRoute(p_idrp_rt);
		return(0);
		}
	return(p_idrp_rt);

}
