#include "include.h"
#include "idrp.h"

/* Routing table code for IDRP */

/* This idrp routing table code attempts to look like the bgp-4 */
/* code in gated.  Since we are doing simular functions, this  */

/* imply better maintenance or simply common confusion.  */
/* idrp specific mainteance routines: */

/* hash table routines
 *  idrp_rid_hash(peer,id) - Sue
 *  - routine to hash the route id and
 *    return an idrpRoute entry
 *
 *  idrp_add_hash_entry(peer,id,p_rt) - Sue
 *  - routine to add new entry to hash table
 *  - give route_id and idrpRoute structure
 *
 *  idrp_free_hash_entry(peer,id) - Sue
 *   - routine to release hash entry from
 *     hash table given id, and relink
 *     the hash table list 
 *
 * idrp_release_hash_tbl(peer) - sue
 *  -  routine to release all of hash table
 *     entries and return 
 *
 * idrp_init_hash_tbl(peer) - sue 
 * - routine to initialize hash table
 *
 *  %% idrpRoute structure handling routines
 * 
 * 1) idrp_alloc_idrpRoute(peer,attr_rec,family)  (Sue's/joint) 
 *     idrpPeer *peer - peer structure pointer which
 *                      route is associated with 
 *     attr_rec - attribute record for route
 *     family - AF.ISO, AF.INET
 *     return pointer to idrpRoute structure 
 *
 * 2) idrp_free_idrpRoute(p_rt)
 *    p_rt - idrp route structure
 * 
 *
 * 3) idrp_free_idrpRoute_entry(id,route_entry_list,circular)
 *     id - id of route
 *     idrpRoute_list
 *     circular - whether list is circular or not
 * 	
 *	- searches list for route_entry_list 
 * 
 * 4) idrp_add_idrpRoute_entry(id,route_entry_list) - 
 *     id - id of route
 *     idrpRoute_list
 *     circular - circular list or not
 *		- creates a new route_id entry and links to list
 * 
 *  
 * idrp_free_outlist  - free an idrpRoute from outlist and relink
 *			outlist around the idrp Route
 */


idrp_rid_hash *idrp_rid_hash_get(peer,id)
idrpPeer *peer;
u_int id;
{
idrp_rid_hash *p_rid;

	p_rid = &peer->hash_table[IDRP_ROUTE_ID_HASH(id)]; 

	/* walk the hash table list */
	while (p_rid)
		{
		if (p_rid->route_id == id)
			{
			return (p_rid);
			}
		p_rid = p_rid->next;
		}
	return ((idrpRoute *) 0);
} 
			

/* idrp_add_hash_entry
 * - add entry to peer hash table for route_id_in
 */

void idrp_add_hash_entry(peer,id,p_route)
idrpPeer *peer;
u_int id;
idrpRoute *p_route[];
{
idrp_rid_hash	*p_rid = 0;
idrp_rid_hash	*p_rid_last = 0;
int		i;
 
	p_rid = &peer->hash_table[IDRP_ROUTE_ID_HASH(id)];

	/* if not the initial entry - then
	 * allocate a entry for this hash entry
	 */
 
	if (p_rid->route_id != IDRP_INIT_ROUTE_ID)
		{
		while (p_rid)
			{
			p_rid_last = p_rid;
			p_rid = p_rid->next;
			}
		p_rid = (idrp_rid_hash *) task_mem_malloc(peer->task,sizeof(idrp_rid_hash));
		p_rid->next = (idrp_rid_hash *) 0;
		p_rid_last->next = p_rid;
		}

	p_rid->route_id = id;
	NLRI_FAMILY_LOOP(i)
		{
		p_rid->p_idrp_rt[i] = p_route[i]; 
		} NLRI_FAMILY_LOOP_END
	return;
}

void idrp_free_hash_entry(peer,id)
idrpPeer *peer;
u_int id;
{
idrp_rid_hash *p_rid = 0;
idrp_rid_hash *p_rid_last = 0;
 
	p_rid = &peer->hash_table[IDRP_ROUTE_ID_HASH(id)];

	/* walk the hash table list looking for id */

	while (p_rid)
		{
		if (p_rid->route_id == id)
			{
			if (p_rid_last)
				{
				/* not first on list */
				/* link over this id */
	
				p_rid_last->next = p_rid->next;
				task_mem_free(peer->task,p_rid);
				return;
				}	
			else
				{
				/* first on list */

				peer->hash_table[IDRP_ROUTE_ID_HASH(id)].next = p_rid->next;
				task_mem_free(peer->task,p_rid);
				return;
				}
			}
		p_rid_last = p_rid;
		p_rid = p_rid->next;
		}
	trace (TR_IDRP,0,"hash entry route id %d for peer %s tried to be freed, but not there ", \
			id,peer->name);

	return;	
}
			
 
void idrp_release_hash_tbl(peer)
idrpPeer *peer;
{
idrp_rid_hash	*p_rid;
idrp_rid_hash	*p_rid_last;
int		i;

	for (i = 0; i < IDRP_HASH_TABLE_SIZE; i++)
		{
		if (peer->hash_table[i].route_id != IDRP_INIT_ROUTE_ID)
			{	
			p_rid = peer->hash_table[i].next;
			while (p_rid)
				{
				/* walk hash chain freeing up the task table */
				p_rid_last = p_rid;
				p_rid = p_rid->next;
				task_mem_free(peer->task,p_rid_last);
				}
			}
		}

	return;
}

void idrp_reinit_hash_tbl(peer)
idrpPeer	*peer;
{
	idrp_release_hash_tbl(peer);
	idrp_init_hash_tbl(peer);
}		

void idrp_init_hash_tbl(peer)
idrpPeer	*peer;
{
int	i;

	for (i = 0; i < IDRP_HASH_TABLE_SIZE; i++)
		{
		peer->hash_table[i].route_id = IDRP_INIT_ROUTE_ID;
		}
}
		
			
/* %% idrpRoute handling routines */

		
/*
 * idrp_alloc_idrpRoute(peer,attr_rec,family)  (Sue's/joint) 
 *     idrpPeer *peer - peer structure pointer which
 *                      route is associated with 
 *     attr_rec - attribute record for route
 *     family - AF.ISO, AF.INET
 *     return pointer to idrpRoute structure 
 */
idrpRoute *
idrp_alloc_idrpRoute(peer,attr_rec,family,route_id_in)
idrpPeer *peer;
idrp_attribute_record *attr_rec;
int family;
int route_id_in;
{
idrpRoute	*p_route;	/* idrp route pointer */

/* allocate the buffer for the idrp route information */
/* this buffer goes in the protocol specific portion  */
/* of the gated route_entry   */
 
        p_route = (idrpRoute *) task_mem_malloc(peer->task, sizeof(idrpRoute));
	bzero(p_route,sizeof(idrpRoute));
        p_route->peer=peer;
        p_route->p_attr = attr_rec;
        p_route->family = family;
	p_route->route_id_in = route_id_in;
        return(p_route);
}


void idrp_free_idrpRoute(p_rt)
idrpRoute *p_rt;
{
	/* assumes the idrpRoute is unlinked from all structures */

	task_mem_free(p_rt->peer->task,p_rt);

}
/* %% idrpRoute_entry handling routines */


/* 
 *  idrp_free_idrpRoute_entry(id,route_entry_list,circular)
 *     id - id of route
 *     idrpRoute_list
 * 	
 *	- searches list for route_entry_list 
 * 	returns - head of list  
 */

idrpRoute_entry *idrp_free_idrpRoute_entry(id,peer,p_rt_list,circular)
int		id;		/* id of route entry */
idrpPeer	*peer;		/* peer route id came from */
idrpRoute_entry *p_rt_list;	/* route list of idrpRoutes */
int circular;			/* TRUE or FALSE */
{
idrpRoute_entry	*p_rt;		/* pointer for walking route list */ 

	if (!p_rt_list)
		{
		trace (TR_IDRP,0,"big time error in idrp no idrp_rt_list to idrp_free_idrpRoute");
		return;
		}

	/*  check to see if head - will happen most of time */ 

	if (p_rt_list->route_id == id)
		{
		/* route id is at head of list */
		
		p_rt = p_rt_list->p_next;
		if (circular)
			{
			/* check to see that this was not */
			/* the only one on the list */

			if (p_rt == p_rt_list)
				{
				/* yes only one on circular list */
				p_rt == NULL;
				}	  
			}
		task_mem_free(peer->task,p_rt_list);
		return (p_rt);
		}

	if (circular)
		{
		return(circular_free_idrpRoute(id,p_rt_list));
		}
	else
		{
		return (free_idrpRoute_ent(id,peer,p_rt_list));
		}
}

/*
 * free_idrpRoute_ent(id,peer,p_rt_list)
 *
 *  free the idrp route entry off a linked list 
 */

idrpRoute_entry *free_idrpRoute_ent(id,peer,p_rt_list)
int 		id;
idrpPeer	*peer;
idrpRoute_entry	*p_rt_list;
{
idrpRoute_entry	*p_rt;		/* pointer for walking route list */ 
idrpRoute_entry	*p_rt_last;	/* pointer for walking route list */ 

	p_rt = p_rt_list->p_next;
	p_rt_last = p_rt_list; 

	/* non-circular list - try for just one */

	if (p_rt == 0)  
		{
		/* one entry on linked list and not there */

		trace (TR_IDRP,0,"no match found list %x route id %d",p_rt_list,id);
		return (p_rt_list);
		}

	/* walk through list looking for id */

	while (p_rt)
		{
		if (p_rt->route_id == id && p_rt->peer == peer)
			{
			/* found the node to free */
			/* relink list */			
	
			p_rt_last->p_next = p_rt->p_next;
			task_mem_free(peer->task,p_rt);
			return (p_rt_list);	
			}
		p_rt_last = p_rt;
		p_rt = p_rt->p_next;
		}

	/* if get here, no match found */

	trace (TR_IDRP,0,"no match found list %x route id %d",p_rt_list,id);
	return (p_rt_list);
}

/*
 * circular_free_idrpRoute(id,p_rt_list)
 *
 *  free the idrp route entry off a linked list 
 */

idrpRoute_entry *circular_free_idrpRoute(id,p_rt_list)
int id;
idrpRoute_entry	*p_rt_list;	/* pointer to route list */ 
{
idrpRoute_entry	*p_rt;		/* pointer for walking route list */ 
idrpRoute_entry	*p_rt_last;	/* pointer for walking route list */ 

	p_rt = p_rt_list->p_next;
	p_rt_last = p_rt_list; 

	/* walk through circular list until we get to begining */

	while (p_rt != p_rt_list)
		{
		/* check to see if ID found */
		if (p_rt->route_id == id)
			{
			p_rt_last->p_next = p_rt->p_next;
			task_mem_free(p_rt->peer->task,p_rt);
			return (p_rt_list);
			}
		p_rt_last = p_rt;
		p_rt = p_rt->p_next;
		}
	
	/* no id found */
	trace (TR_IDRP,0,"no match found for route_id circular list %x %d",p_rt_list,id);
	return (p_rt_list);
		
}

/* 
 * 6.) idrp_add_idrpRoute_entry(id,route_entry_list) - 
 *     id - id of route
 *     idrpRoute_list
 */

idrpRoute_entry *idrp_add_idrpRoute_entry(id,p_rtl)
u_int id;
idrpRoute_entry *p_rtl;	/* pointer to route list */
{
idrpRoute_entry *p_rt;

	/* walk to end of the linked list (not circular) */ 

	p_rt = p_rtl;
	while (p_rt->p_next)
		{
		p_rt = p_rt->p_next;
		}
	p_rt->p_next = (idrpRoute_entry *) task_mem_malloc(p_rt->peer->task,sizeof(idrpRoute_entry));
	bzero(p_rt->p_next,sizeof(idrpRoute_entry));
	p_rt->p_next->route_id = id;
	p_rt->p_next->mem_type = IDRP_TASK_MEM;
	return(p_rt->p_next);
}

int
idrp_rt_within_RD(p_idrp_rt)
idrpRoute	*p_idrp_rt;
{
	if (p_idrp_rt->route_id_in == IDRP_LOCAL_ROUTE_ID)
		{
		return(TRUE);
		}

	if (p_idrp_rt->route_id_in == IDRP_EXT_INFO_ROUTE_ID)
		{
		if (p_idrp_rt->p_attr->local_mask & IDRP_LOCAL_EXT_INFO)
			{
			return (TRUE);
			}

		if (p_idrp_rt->p_attr->idrp_local_proto == RTPROTO_BGP)
			{
			return(idrp_bgp_within_rd(p_idrp_rt));
			}
		}
	return(FALSE);
}

 /*
 * idrp_free_outlist - relink an idrpRoute's Route_out list
 *		    around a route
 * 
 */

void 
idrp_free_outlist(p_idrp_rt)
idrpRoute	*p_idrp_rt;
{
int		i;

	PEER_CNT_LOOP(i)
                {
                if (p_idrp_rt->route_out[i].route_id > 0)
                        {
			if (!p_idrp_rt->route_out[i].p_next)
				{
				/* only one route on the output route id */
				p_idrp_rt->route_out[i].route_id = 0;
				}
			else
				{
				if (p_idrp_rt->route_out[i].p_next == p_idrp_rt)
					{
					/* Only one link on the route-out
					 *  clear the route_id and the p_next 
					 * field for this _id
					 */
					p_idrp_rt->route_out[i].p_next = (idrpRoute *) 0;
					p_idrp_rt->route_out[i].route_id = 0;
					}
				else
					{
					p_idrp_rt->route_out[i].p_next = p_idrp_rt->route_out[i].p_next->route_out[i].p_next;
					}
				}
			}
		} PEER_CNT_LOOP_END;
}
